/* Operservice 2
 * Definition of global variables, datatypes etc.
 */

#ifndef GLOBALS_H

#define GLOBALS_H

#include "mysql_version.h"
#include "mysql.h"
#include "config.h"

#define operservversion "2.20"

#define um_o 1     // +o on a channel (op)
#define um_v 2     // +v on a channel (voice)
#define cm_p 1     // chanmode +p (private)
#define cm_s 2     // chanmode +s (secret)
#define cm_m 4     // chanmode +m (moderated)
#define cm_n 8     // chanmode +n (no outside message)
#define cm_t 16    // chanmode +t (only ops can change topics)
#define cm_i 32    // chanmode +i (invite only)
#define cm_l 64    // chanmode +l (limit)
#define cm_k 128   // chanmode +k (key)

#define NICKLEN 15     // Maximum nick length
#define USERLEN 10     // Maximum ident length
#define HOSTLEN 63     // Maximum hostname length
#define REALLEN 50     // Maximum realname length
#define CHANNELLEN 200 // Maximum length of a channel name
#define SRVLEN 200     // Maximum length of a servername
#define TOPICLEN 250   // Maximum length of a topic
#define CHANKEYLEN 26  // Maximum length of a channelkey
#define MAXUMODES 20   // Maximum number of usermodes
#define AUTHUSERLEN 40 // Maximum Length of a Username for authing
#define AUTHPASSLEN 40 // Maximum length of a Password for authing
#define RNGLEN 150     // Realname GLINE length (for the mask)
#define RNGREAS 100    // Maximum length of reason in Realname-GLINEs
#define RNGTIM 1800    // How many seconds the realname-glines will be set
#define SIZEOFUL 15000 // How many chained lists form the Userlist
#define SIZEOFCL 15000 // The same for the channellist
#define SIZEOFNL 10000  // and the Nicklist
#define TMPSSIZE 1040  // Size of temporary strings. this is 1040 because a
                       // string from the server has at most 512 bytes.
                       // 2 * 512 + some bytes for safety makes 1040
#define SRVSHIFT 18    // Servernumeric-Multiplier. Servernumeric SHIFTLEFT
                       // SRVSHIFT + Clientnumeric = complete client numeric
#define SRVNUMMULT 262144 // The same, but for use with "%"
#define ARLIM1 100     // If new memory is needed for an array, this many
                       // records will be allocated.
#define ARLIM2 150     // If more than this number of records are free in an
                       // array, memory is freed. Those are only default
                       // values, arrays.c provides functions to change them.
#define MAXPARAMS 30   // Every commandline received from the server gets split
                       // up now. I'm pretty sure this will never exceed 30
                       // parts.
#define MODNAMELEN 50  // Maximum length of a module name (without path)
#define MODDESCLEN 100 // Length for short module description.
#define COMMANDLEN 50  // Max. length for dynamically loaded commands

#define NOTICEMASKDEFAULT 1    // Default noticemask for ircops
#define NM_GLINE 1     // GLINEs set / removed
#define NM_RNGL  2     // Realname-GLINEs set

#define TRUSTNAMELEN 70     // Length of the name for a group of trusted hosts
                            // - MUST be at least HOSTLEN!
#define TRUSTCONTACTLEN 150 // Length of admin-email
#define TRUSTCOMLEN 200     // Length of the field for additional comments
#define SIZEOFHL 1000  // Size of hostlist (for clonetracking)
#define SIZEOFTL 200   // Size of trustlist (contains trusted hosts)
#define SIZEOFIDMAP 200     // Size of the table mapping trustgroup-ids to names

typedef struct {
  unsigned long cursi;
  unsigned long maxsi;
  unsigned int memberlen;
  unsigned int arlim1;
  unsigned int arlim2;
  void *content;
} array;

typedef struct {
  char nick[NICKLEN+1];
  int hopsaway;
  long numeric;
  char ident[USERLEN+1];
  char host[HOSTLEN+1];
  char umode[MAXUMODES+1];
  char realname[REALLEN+1];
  array chans;
  unsigned long realip;     /* The real IP of the client as a 32 bit integer */
                            /* Obviously this is only compatible with IPv4! */
  unsigned long connectat;  /* Timestamp: When this client connected */
  struct userdata *next;    /* Pointer to the next user (linked list) */
} userdata; /* This struct holds information about a user on the net */

typedef struct {
  char nick[SRVLEN+1];
  int hopsaway;
  long numeric;
  long createdat;
  unsigned long connectat;
  long connectedto;
  long maxclients;
  long pingreply;
  long pingsent;
  char ports[1000];
  struct serverdata *next;
} serverdata;

typedef struct {
  long numeric;
  int flags;
} chanuser;

typedef struct {
  char banmask[260];
  struct chanban *next;
} chanban;

typedef struct {
  char name[CHANNELLEN+1];
  long createdat;
  int flags;
  char chankey[CHANKEYLEN+1]; // Channelkey - if the channel is not +k, the content
                              // of this is undefined!!
  int chanlim;        // Channellimit - same as with chankey
  array users;
  chanban *bans;      // Not yet used
  char topic[TOPICLEN+1];
  struct channel *next;
} channel;

typedef struct {
  char c;
  int v;
} flags;

typedef struct {
  unsigned int a;
  unsigned int b;
  unsigned int c;
  unsigned int n;  // Number of clients on that subnet
} subnetdata;

typedef struct {
  long maxsize;
  long cursize;
  subnetdata *s;
} subnetlist;

typedef struct {
  unsigned int a;
  unsigned int b;
  unsigned int c;
  unsigned int n;  // For how many connects this subnet is trusted
  unsigned long expires;
  char reason[256];
} trustedsubnet;

typedef struct {
  char username[AUTHUSERLEN+1];
  char password[AUTHPASSLEN+1]; // if authlevel is >=100, this is only a hash, not the
                     // cleartext password!
  int authlevel;     // 1 - 1000
  long lastauthed;
  int wantsnotice;
  unsigned long noticemask;
} authdata;

typedef struct {
  long numeric;
  char authedas[AUTHUSERLEN+1];
  int authlevel;
  int wantsnotice;
  unsigned long noticemask;
  struct autheduser *next;
} autheduser;

typedef struct {
  char r[REALLEN+1];
  int n;
} rnlentry;

typedef struct {
  char mask[RNGLEN+1];
  char creator[AUTHUSERLEN+1];
  unsigned long expires;
  int howtogline; /* 1: *@host
                     2: user@host
                  */
  char reason[RNGREAS+1];
  unsigned long timesused;  // Only some little stats, how often this has been
                            // used since O was started
} realnamegline;

typedef struct {
  char nick[NICKLEN+1];
  long numeric;
  char ident[USERLEN+1];
  char host[HOSTLEN+1];
  char realname[REALLEN+1];
  unsigned long connectat;
} afakeuser;

typedef struct {
  char name[SRVLEN+1];
  int state;
  unsigned long time;
} splitserv;

typedef struct {
  char name[MODNAMELEN+1];
  void *handle;
  char shortdesc[MODDESCLEN+1];
} loadedmod;

typedef struct {
  char name[COMMANDLEN+1];
  char providedby[MODNAMELEN+1];
  char *hlptxt;
  int operonly;
  void (*func)(long unum, char *tail);
} dyncommands;

typedef struct {
  unsigned long id; // MUST NOT be 0
  char name[TRUSTNAMELEN+1];
  int trustedfor;
  int currentlyon;
  unsigned long expires;
  int maxperident;
  int enforceident;
  char contact[TRUSTCONTACTLEN+1];
  char comment[TRUSTCOMLEN+1];
  array identcounts;
  struct trustedgroup *next;
} trustedgroup;

typedef struct {
  char hostname[HOSTLEN+1];
  unsigned long id; // MUST NOT be 0
  struct trustedhost *next;
} trustedhost;

typedef struct {
  char ident[USERLEN+1];
  int currenton;
} identcounter;

typedef struct {
  char hostname[HOSTLEN+1];
  int currenton;
} hostcounter;

extern serverdata *sls;
extern userdata * uls[SIZEOFUL]; /* Userlist */
extern channel * cls[SIZEOFCL];  /* Channellist */
extern array nls[SIZEOFNL];      /* Nick-lists */
extern autheduser *als;  /* List of authed users */
extern flags chanflags[];
extern int numcf; /* Number of Chanflags */
extern flags userflags[];
extern int numuf; /* Number of Userflags (Flags of a user on a channel) */
extern FILE *sockin, *sockout;
extern char lastline[520]; /* Last line received from the ircserver */
extern char tokens[];
extern unsigned int reversetok[];
extern char crytok[];
extern char upwdchars[];
extern char command[TMPSSIZE];
extern char conpass[100];
extern char mynick[NICKLEN+1];
extern char myname[100];
extern char conto[100];
extern unsigned int conport;
extern char sqluser[100];
extern char sqlpass[100];
extern char sqlhost[100];
extern char sqldb[100];
extern int sqlport;
extern char servernumeric[3];
extern char uplnum[3];
extern char lastburstchan[520];
extern int lastburstflags;
extern long starttime;  // Timestamp, contains starttime of the service
extern int burstcomplete;  // Is set to 1 once the netburst completed
extern char logline[2500];
extern char logfile[FILENAME_MAX];
extern char helppath[300];
extern long timestdif;   // Timestamp-difference
extern MYSQL sqlcon;
extern array mainrnl[256];
extern subnetlist mainsnl[256];
extern long snlstartsize;
extern trustedsubnet *tsn;
extern long tsnsi;
extern long tsncu;
extern array rngls;
extern int mfwarn;
extern long lastsettime;
extern long waitingforping;
extern long lasthourly;
extern long lastfakenum;
extern long resyncglinesat;
extern char sendglinesto[20480];
extern char *noreplytoping;
extern array myfakes; /* List of my fakeusers */
extern char *nicktomsg;
extern char params[MAXPARAMS][TMPSSIZE]; // The last line received from the
                                         // server, split at the spaces.
extern int paramcount;                   // How many slots of params[] are used
extern char sender[TMPSSIZE];            // The sender of the last line -
                                         // always as a numeric!
extern array splitservers; // List of splits, see splitdb.c
extern char modpath[FILENAME_MAX]; // Path where loadable modules reside.
extern char modend[10];        // What filesuffix modules have
extern array modulelist;      // List of all loaded modules
extern array commandlist;     // List of all loaded commands
extern int defaultclonewarn;  // Default for warning about clones (6 on quakenet, 0 to disable)
extern trustedhost * trustedhosts[SIZEOFTL];       // The trusted hosts
extern array untrustedhosts[SIZEOFHL];     // normal, untrusted hosts
extern trustedgroup * trustedgroups[SIZEOFIDMAP];

/* End of global variable definitions */

/* Define important functions */

/* Operservice.c */
int alreadyinglinesto(char *nick);
void showhelp(long unum, char *command);
void getuserdata(authdata *retad, char *username);
long countusersinul();
void createrandompw(char *a,int n);
int updateuserinul(authdata ad);
autheduser * getalentry(long unum);
void newalentry(long unum);
void addautheduser(long unum, authdata ad);
void delautheduser(long unum);
int opensql(char *host, char *username, char *password, char *database, int port);
int tcpopen(char *host, unsigned int port);
int numericinuse(long num);
void delchan(char *name);
void deluserfromallchans(long num);
void delallchanson(long num);
channel * getchanptr(char *name);
long getsrvnum(char *a);
int serverdoesnotreplytopings(serverdata *s);
void pingallservers();
long countusershit(char *pattern);
void noticeallircops(char *thetext);
void delalluserson(long num);
void delsrv(char *a);
void deluserfromchan(char *name, long numeric);
void newserver(char *nick, int hopsaway, long numeric, long createdat, long connectat, long connectedto, long maxclients);
int chanexists(char *name);
channel * newchan(char *name, long createdat);
void setchankey(char *name, char *key);
void setchanke2(channel *a, char *key);
void saveall();
long countusersonsrv(long num);
long countusersonthenet();
long nicktonum(char *a);
void numtohostmask(long num, char *res);
void numtonick(long num, char *res);
int loadconfig(char *filename);
void changechanmode(char *name, long numeric, int sm, int flags);
void changechanmod2(channel *a, long numeric, int sm, int flags);
int getchanmodes(char *name, long numeric);
int getchanmode2(channel *a, long numeric);
long getmaxping();
int gotallpingreplys();
void purgeglines();
void remgline(char *hostmask, int sendtonet);
void addgline(char *hostmask, char *reason, char *bywhom, long howlong, int sendtonet);
void addusertochan(char *name, long numeric);
void addusertocptr(channel *b, long numeric);
void delchanfla2(channel *a, int flags);
void setchanfla2(channel *a, int flags);
void delchanflag(char *name, int flags);
void setchanflag(char *name, int flags);
void setchanli2(channel *a, int lim);
void setchanlim(char *name, int lim);
void addnicktoul(userdata *a);
int rnladd(char *realname);
void rnldel(char *realname);
long nicktonu2(char *a);
void dochanneljoins(void);
void addchantouser2(long num, channel *c);
void addchantouser(userdata *a, channel *c);
void delchanfromuser(long num, char *channam);
void addtonicklist(userdata *a);
void delfromnicklist(char *nick);
/* End of Operservice.c */

/* subnetlist.c */
int snlcreate(subnetlist *snl, long mysize);
void snlfree(subnetlist *snl);
int snlcomp(subnetdata *a, subnetdata *b);
void snlsort(subnetlist *snl);
unsigned int snladd(subnetlist *snl, unsigned long IP);
void snldel(subnetlist *snl, unsigned long IP);
/* End of subnetlist.c */

/* fakeuser.c */
void createfakeuser(long num, char *nick, char *ident, char *host, char *realname, int sendtonet);
int clinuminuse(long senu, long clnu);
long getfreefakenum();
void fakeuseradd(char *nick, char *ident, char *host, char *realname, long numeric);
void fakeuserdel(long numeric);
void fakeuserkill(long numeric);
void fakeuserload();
void fakeusersave();
void dofakelist(long unum, char *tail);
void dofakekill(long unum, char *tail);
void docreatefakeuser(long unum, char *tail);
/* End of fakeuser.c */

/* realnamegline.c */
void rnglcreate();
void rngladd(char *mask, char *creator, long expires, int howtoban, char *reason);
void rnglcheck(userdata *a);
void rngldel(char *mask);
void rnglexpire();
void rnglload();
void rnglsave();
void dorngline(long unum, char *tail);
void dornungline(long unum, char *tail);
void dornglist(long unum, char *tail);
/* End of realnamegline.c */

/* subnettrust.c */
void createtl();
void snltrustadd(unsigned int a, unsigned int b, unsigned int c, unsigned int n, unsigned long expires, char *reason);
int snltrustdel(unsigned int a, unsigned int b, unsigned int c);
unsigned int getsnltrust(unsigned long IP);
void trustlistexpire();
void loadtrustlist();
void savetrustlist();
/* End of subnettrust.c */

/* general.c */
time_t getnettime();
void splitip(unsigned long IP, unsigned int *a, unsigned int *b, unsigned int *c, unsigned int *d);
unsigned int clhash(char *txt);
unsigned int ulhash(long num);
unsigned int nlhash(char *nick);
void encryptpwd(char *res, char *pass);
int checkpwd(char *encrypted, char *pass);
int match2strings(char *patrn, char *strng);
int isflagset(int v, int f);
int alldigit(char *s);
unsigned long tokentolong(char* token);
void longtotoken(unsigned long what, char* result, int dig);
unsigned long iptolong(unsigned int a, unsigned int b, unsigned int c, unsigned int d);
void toLowerCase(char* a);
void normnum(char* a);
void delchar(char *a, char b);
void appchar(char *a, char b);
long durationtolong(char *dur);
void longtoduration(char *dur, long what);
int ischarinstr(char a, char *b);
unsigned long crc32(const unsigned char *buf);
void somethingtonumeric(char *what, char *result);
void unescapestring(char *src, char *targ);
userdata * getudptr(long numeric);
int killuser(long num);
void mystrncpy(char *targ, char *sour, long n);
void putlog();
int getauthlevel(long unum);
void getauthedas(char *res, long unum);
int getwantsnotice(long unum);
void msgtouser(long unum, char *txt);
int checkauthlevel(long unum, int minlevel);
void sendtonoticemask(unsigned long mask, char *txt);
int isircop(long num);
unsigned long getnoticemask(long unum);
void setnoticemask(long unum, unsigned long mask);
/* End of general.c */

/* serverhandlers.c */
void handlenickmsg(void);
void handlequitmsg(void);
void handleservermsg(int firsts);
void handleping(void);
void handlekillmsg(void);
void handlesquit(void);
void handlemodemsg(void);
void handleprivmsg(void);
void handleburstmsg(void);
void handlecreatemsg(void);
void handlejoinmsg(void);
void handlepartmsg(void);
void handlekickmsg(void);
void handleglinemsg(void);
void handlestatsPreply(void);
void handlestatsureply(void);
void handlestatsmsg(void);
void handleeoback(void);
void handletopic(void);
/* End of serverhandlers.c */

/* usercommands.c */
void dowhois(long unum, char *tail);
void dostatus(long unum, char *tail);
void doopchan(long unum, char *tail);
void dokickcmd(long unum, char *tail);
void dojupe(long unum, char *tail);
void dooperhelp(long unum, char *tail);
void douserhelp(long unum, char *tail);
void dochancmd(long unum, char *tail);
void doresync(long unum, char *tail);
void dosemiresync(long unum, char *tail);
void dodeopall(long unum, char *tail);
void doglist(long unum, char *tail);
void dogline(long unum, char *tail);
void doungline(long unum, char *tail);
void docountusers(long unum, char *tail);
void domfc(long unum, char *tail);
void dosettimecmd(long unum, char *tail);
void doserverlist(long unum, char *tail);
void dohello(long unum, char *tail);
void doauth(long unum, char *tail);
void dochangepwd(long unum, char *tail);
void dochangelev(long unum, char *tail);
void dodelusercmd(long unum, char *tail);
void donoticeset(long unum, char *tail);
void dobroadcast(long unum, char *tail);
void dosbroadcast(long unum, char *tail);
void dosyncglines(long unum, char *tail);
void dosubnetlist(long unum, char *tail);
void dolistauthed(long unum, char *tail);
void dosave(long unum, char *tail);
void dosntrustadd(long unum, char *tail);
void dosntrustdel(long unum, char *tail);
void dosntrustlist(long unum, char *tail);
void dospew(long unum, char *tail);
void dornc(long unum, char *tail);
void dointtobase64(long unum, char *tail);
void dobase64toint(long unum, char *tail);
void dodie(long unum, char *tail);
void donoticemask(long unum, char *tail);
/* End of usercommands.c */

/* splitdb.c */
void addsplit(char *name);
void changesplitstatus(char *name, int newstatus);
int getsplitstatus(char *name);
void delsplit(char *name);
void clearsplits(void);
void splitpurge(void);
void dosplitlist(long unum, char *tail);
void dosplitdel(long unum, char *tail);
/* End of splitdb.c */

/* arrays.c */
void array_init(array *a, unsigned int memberlen);
void array_setlim1(array *a, unsigned int lim);
void array_setlim2(array *a, unsigned int lim);
unsigned long array_getfreeslot(array *a);
void array_delslot(array *a, unsigned long slotn);
void array_free(array *a);
/* End of arrays.c */

/* dynamic.c */
int loadandinitmodule(char *name);
int isloaded(char *name);
int removemodule(char *name);
void setmoduledesc(char *name, char *txt);
void registercommand(char *mod, char *name, void *func, int operonly, char *hlptxt);
void deregistercommand(char *name);
int dodyncmds(char *cmd, long unum, char *tail, int oper);
void dyncmdhelp(long unum, int oper);
void doloadmod(long unum, char *tail);
void dounloadmod(long unum, char *tail);
void doreloadmod(long unum, char *tail);
void dolsmod(long unum, char *tail);
/* End of dynamic.c */

/* trusts.c */
unsigned int hlhash(char *host);
unsigned int tlhash(char *host);
unsigned int tgshash(unsigned long ID);
unsigned long istrusted(char *host);
int addtountrustedhost(char *host);
void delfromuntrustedhost(char *host);
void trustnewclient(char *ident, char *host);
void trustdelclient(char *ident, char *host);
void dotrustgroupadd(long unum, char *tail);
void dotrustgroupdel(long unum, char *tail);
void dotrustadd(long unum, char *tail);
void dotrustdel(long unum, char *tail);
trustedgroup * findtrustgroupbyID(unsigned long ID);
trustedgroup * findtrustgroupbyname(char *name);
unsigned long getfreetgid();
int destroytrustgroup(unsigned long id);
void trustgroupexpire(void);
unsigned long loadtrusts(void);
void savetrusts(void);
void dotrustlist(long unum, char *tail);
void dotrustgline(long unum, char *tail);
void dotrustspew(long unum, char *tail);
/* End of trusts.c */

/* End of important functions defining */

#endif
