/*
Operservice 2 - serverhandlers.c
This baby handles all commands received from the server.
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "mysql_version.h"
#include "mysql.h"
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "globals.h"

void handlenickmsg() {
  unsigned long tmpl;
  int slen; userdata *a;
  slen=strlen(sender);
  if (slen==5) {
    // Nickchange
    if (paramcount!=4) {
      sprintf(logline,"!!! Nickchange failed (wrong # args): %s !!!",lastline); putlog();
      return;
    }
    toLowerCase(params[2]);
    tmpl=tokentolong(sender);
    a=uls[ulhash(tmpl)];
    while (a!=NULL) {
      if (a->numeric==tmpl) {
        if (strlen(params[2])>NICKLEN) {
          sprintf(logline,"!!! handlenickmsg/ren: Nick exceeds NICKLEN: %s !!!",lastline); putlog();
        } else {
          delfromnicklist(a->nick);
          strcpy(a->nick,params[2]);
          addtonicklist(a);
        }
        a->connectat=strtoul(params[3],NULL,10);
        return;
      }
      a=(void *)a->next;
    }
    sprintf(logline,"!!! Nickchange failed (no such user): %s !!!",lastline); putlog();
  }
  if (slen==2) {
    // New Nick introduced to the network
    int ipa, ipb, ipc, ipd; char realname[TMPSSIZE]; int rnstart; char *c;
    char umode[TMPSSIZE]; char numtmp[TMPSSIZE]; char realiptmp[TMPSSIZE];
    int res;
    if (paramcount<10) {
      sprintf(logline,"!!! Nick introduction failed (wrong # args): %s !!!",lastline); putlog();
      return;
    }
    if (params[7][0]=='+') { rnstart=10; } else { rnstart=9; }
    if ((rnstart==10) && (paramcount<11)) {
      sprintf(logline,"!!! Nick introduction failed (no realname at all?!): %s !!!",lastline);
      putlog();
      return;
    }
    realname[0]='\0';
    c=&lastline[0];
    while ((*c!='\0') && (*c!=':')) { c++; }
    ipa=0; c++;
    for (;(*c!='\0');c++) { realname[ipa]=*c; ipa++; }
    realname[ipa]='\0';
    if (rnstart==9) { strcpy(umode,"+"); } else { strcpy(umode,params[7]); }
    if (rnstart==9) { strcpy(numtmp,params[8]); } else { strcpy(numtmp,params[9]); }
    if (rnstart==9) { strcpy(realiptmp,params[7]); } else { strcpy(realiptmp,params[8]); }
    normnum(numtmp);
    tmpl=tokentolong(realiptmp);
    toLowerCase(params[2]);
    if (nicktonu2(params[2])>=0) { /* Nick collission */
      long thisisnotgood; long ting2; userdata *ting3;
      thisisnotgood=nicktonu2(params[2]);
      ting2=strtoul(params[4],NULL,10);
      ting3=getudptr(thisisnotgood);
      if (ting3==NULL) {
        sprintf(logline,"!!! Nick collission: Fucked up hashtable?! !!!");
        putlog(); return;
      }
      if ((strcmp(params[5],ting3->ident)==0) && (strcmp(params[6],ting3->host)==0)) {
        /* This guy collided itself, which means both clients get killed */
        sprintf(logline,"!!! Nick collission: %s collided itself !!!",ting3->nick);
        putlog();
        deluserfromallchans(thisisnotgood);
        delautheduser(thisisnotgood);
        if (!killuser(thisisnotgood)) {
          fakeuserkill(thisisnotgood);
          sprintf(logline,"!!! KILL failed on nickcollide: %s !!!",lastline); putlog();
        }
        return;
      }
      if (ting2==ting3->connectat) {
        /* We cannot decide who was here earlier, so we have to kill both */
        sprintf(logline,"!!! Nick collission: %s - same timestamp !!!",ting3->nick);
        putlog();
        deluserfromallchans(thisisnotgood);
        delautheduser(thisisnotgood);
        if (!killuser(thisisnotgood)) {
          fakeuserkill(thisisnotgood);
          sprintf(logline,"!!! KILL failed on nickcollide: %s !!!",lastline); putlog();
        }
        return;
      }
      if (ting2>ting3->connectat) {
        /* Our client gets killed - so no need to add him at all, just return */
        sprintf(logline,"!!! Nick collission: %s overriden by old client !!!",ting3->nick);
        putlog();
        return;
      }
      /* Old client gets killed and replaced by the new one */
      sprintf(logline,"!!! Nick collission: %s overriden by new client !!!",ting3->nick);
      putlog();
      deluserfromallchans(thisisnotgood);
      delautheduser(thisisnotgood);
      if (!killuser(thisisnotgood)) {
        fakeuserkill(thisisnotgood);
        sprintf(logline,"!!! KILL failed on nickcollide: %s !!!",lastline); putlog();
      }
    }
    /* Check for enforced ident on trusted host - in that case KILL immeadiately */
    if (strlen(params[6])>HOSTLEN) {
      sprintf(logline,"!!! Hostname exceeds HOSTLEN: %s !!!",lastline); putlog();
      return;
    }
    if (params[5][0]=='~') {
      unsigned long thid; char tmph[HOSTLEN+1];
      strcpy(tmph,params[6]);
      thid=istrusted(tmph);
      if (thid>0) {
        trustedgroup *tg;
        tg=findtrustgroupbyID(thid);
        if (tg!=NULL) {
          if (tg->enforceident==1) {
            /* Kill the loser! */
            fprintf(sockout,"%s D %s :IDENTD required from your host\r\n",servernumeric,
              numtmp);
            fflush(sockout);
            sprintf(logline,"[Missing IDENTD] Killed %s!%s@%s",params[2],params[5],params[6]);
            putlog();
            return;
          }
        }
      }
    }
    res=snladd(&(mainsnl[tmpl>>24]),tmpl);
    if (res>getsnltrust(tmpl)) {
      if (((res-getsnltrust(tmpl))%10)==1) {
        char tmps2[TMPSSIZE];
        splitip(tmpl,&ipa,&ipb,&ipc,&ipd);
        sprintf(tmps2,"Mirkforce-Warning: %d clients connected from %d.%d.%d.* - CHECK BEFORE YOU GLINE!!!! (use \"subnetlist %d.%d.%d\")",res,ipa,ipb,ipc,ipa,ipb,ipc);
        noticeallircops(tmps2);
      }
    }
    a=(userdata *) malloc(sizeof(userdata));
    a->next=NULL;
    a->hopsaway=atoi(params[3]);
    a->connectat=strtoul(params[4],NULL,10);
    a->realip=tokentolong(realiptmp);
    a->numeric=tokentolong(numtmp);
    if (strlen(params[2])>NICKLEN) {
      sprintf(logline,"!!! Nick exceeds NICKLEN: %s !!!",lastline); putlog();
      free(a); return;
    } else {
      strcpy(a->nick,params[2]);
    }
    if (strlen(params[5])>USERLEN) {
      sprintf(logline,"!!! Ident exceeds USERLEN: %s !!!",lastline); putlog();
      free(a); return;
    } else {
      strcpy(a->ident,params[5]);
    }
    strcpy(a->host,params[6]);
    if (strlen(&umode[1])>MAXUMODES) {
      sprintf(logline,"!!! User has too many Usermodes: %s !!!",lastline); putlog();
      free(a); return;
    } else {
      strcpy(a->umode,&umode[1]);
    }
    if (strlen(realname)>REALLEN) {
      free(a); return;
      sprintf(logline,"!!! Realname exceeds REALLEN: %s !!!",lastline); putlog();
    } else {
      strcpy(a->realname,realname);
    }
    array_init(&(a->chans),sizeof(channel *));
    array_setlim1(&(a->chans),22);
    array_setlim1(&(a->chans),30);
    rnglcheck(a);
    res=rnladd(realname);
    // Add warning if this realname is there too often
    trustnewclient(a->ident,a->host);
    addnicktoul(a);
    addtonicklist(a);
  }
}

void handlequitmsg() {
  long tmpl;
  // We have a QUIT numeric
  tmpl=tokentolong(sender);
  deluserfromallchans(tmpl);
  delautheduser(tmpl);
  if (!killuser(tmpl)) {
    sprintf(logline,"!!! Deluser failed: %s !!!",lastline); putlog();
  }
}

void handleservermsg(int firsts) {
  int xofs; char tmpconnectedto[TMPSSIZE], tmps2[TMPSSIZE];
  if (firsts==1) {
    xofs=0;
    strcpy(tmpconnectedto,servernumeric);
    if (paramcount<7) { sprintf(logline,"!!! Failed to parse Servermessage: %s !!!",lastline); putlog(); return; }
    /* We are the authoritative timesource, so WE set the networktime */
/*    timestdif=strtol(params[4],NULL,10)-getnettime();
    sprintf(logline,"Timestamp-difference: %ld",timestdif); putlog(); */
  } else {
    xofs=1;
    strcpy(tmpconnectedto,sender);
    if (paramcount<8) { sprintf(logline,"!!! Failed to parse Servermessage: %s !!!",lastline); putlog(); return; }
  }
  normnum(params[6+xofs]);
  strcpy(tmps2,params[6+xofs]);
  tmps2[2]='\0';
  if (firsts==1) { strcpy(uplnum,tmps2); }
  newserver(params[1+xofs],atoi(params[2+xofs]),tokentolong(tmps2),
    strtol(params[3+xofs],NULL,10),strtol(params[4+xofs],NULL,10),
    tokentolong(tmpconnectedto),tokentolong(&params[6+xofs][2]));
  if (params[5+xofs][0]=='P') {
//    sprintf(logline,"Server %s already seems to be merged, deleting split",params[1+xofs]);
//    putlog();
    delsplit(params[1+xofs]);
  }
  if (burstcomplete) {
    resyncglinesat=getnettime()+60;
    if (sendglinesto[0]=='\0') {
      strcpy(sendglinesto,params[1+xofs]);
    } else {
      if (!alreadyinglinesto(params[1+xofs])) {
        if (strlen(sendglinesto)<20000) { strcat(sendglinesto,","); strcat(sendglinesto,params[1+xofs]); }
      }
    }
  }  
}

void handleping() {
  static char tmps[TMPSSIZE]="";
  sscanf(lastline,"%s",tmps);
  fprintf(sockout,"%s Z %s\r\n",servernumeric,tmps);
  fflush(sockout);
  sprintf(logline,"Ping? Pong!"); putlog();
}

void handlekillmsg() {
  long num;
  char tmps2[TMPSSIZE];
  if (paramcount<3) {
    sprintf(logline,"!!! Couldnt parse killmessage: %s !!!",lastline); putlog();
    return;
  }
  normnum(params[2]);
  sprintf(tmps2,"%sAAB",servernumeric);
  if (strcmp(params[2],tmps2)==0) {
    longtotoken(iptolong(127,0,0,1),tmps2,6);
    fprintf(sockout,"%s N %s 1 %ld operserv %s +odk %s %sAAB :This is operservice %s\r\n",
      servernumeric,mynick,starttime,myname,tmps2,servernumeric,operservversion);
    dochanneljoins();
    return;
  }
  num=tokentolong(params[2]);
  deluserfromallchans(num);
  delautheduser(num);
  if (!killuser(num)) {
    fakeuserkill(num);
    sprintf(logline,"!!! KILL failed: %s !!!",lastline); putlog();
  }
}

void handlesquit() {
  if (paramcount<3) {
    sprintf(logline,"!!! Failed to parse SQUIT: %s !!!",lastline); putlog();
    return;
  }
  delsrv(params[2]);
}

void handlemodemsg() {
  userdata *a; int t1, t2;
  if (paramcount<4) {
    printf("!!! Failed to parse MODE: %s !!!\n",lastline);
    return;
  }
  if ((params[2][0]!='#') && (params[2][0]!='+') && (params[2][0]!='&')) {
    /* User changes Usermodes - since a user can only change it's own usermode
       and not even services can change usermodes, we can just search for the
       sender (parameter 0) instead of the nick (par 2). */
    long sendnum;
    sendnum=tokentolong(sender);
    a=uls[ulhash(sendnum)];
    while(a!=NULL) {
      if (a->numeric==sendnum) {
        t1=0; t2=1;
        while (params[3][t2]!='\0') {
          if ((params[3][t2]=='+') || (params[3][t2]=='-')) {
            if (params[3][t2]=='+') { t1=1; } else { t1=0; }
          } else {
            if (t1==1) {
              if (strlen(a->umode)<(MAXUMODES-1)) {
                appchar(a->umode,params[3][t2]);
              }
            } else {
              delchar(a->umode,params[3][t2]);
            }
          }
          t2++;
        }
        return;
      }
      a=(void *)a->next;
    }
    sprintf(logline,"!!! Usermodechange failed: %s !!!",lastline); putlog();
  } else {
    // Mode on a channel
    channel *tmpcp; char *c;
    toLowerCase(params[2]);
    tmpcp=getchanptr(params[2]); /* Convert the channelname to a pointer to the 
                                channel data to avoid overhead */
    if (tmpcp==NULL) {
      sprintf(logline,"!!! Modechange for nonexistant channel %s !!!",params[2]); putlog();
      return;
    }
    t2=0; // assume "+"
    c=&params[3][0]; t1=4;
    for (;(*c!='\0');c++) {
      if (*c=='+') { t2=0; }
      if (*c=='-') { t2=1; }
      if (*c=='o') {
        if (paramcount<=t1) {
          sprintf(logline,"!!! Failed to parse modes: %s !!!",lastline); putlog(); return;
        } else {
          normnum(params[t1]);
          if (t2==1) {
            changechanmod2(tmpcp,tokentolong(params[t1]),2,um_o);
          } else {
            changechanmod2(tmpcp,tokentolong(params[t1]),1,um_o);
          }
          t1++;
        }
      }
      if (*c=='v') {
        if (paramcount<=t1) {
          sprintf(logline,"!!! Failed to parse modes: %s !!!",lastline); putlog(); return;
        } else {
          normnum(params[t1]);
          if (t2==1) {
            changechanmod2(tmpcp,tokentolong(params[t1]),2,um_v);
          } else {
            changechanmod2(tmpcp,tokentolong(params[t1]),1,um_v);
          }
        }
      }
      if (*c=='m') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_m);
        } else {
          delchanfla2(tmpcp,cm_m);
        }
      }
      if (*c=='n') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_n);
        } else {
          delchanfla2(tmpcp,cm_n);
        }
      }
      if (*c=='t') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_t);
        } else {
          delchanfla2(tmpcp,cm_t);
        }
      }
      if (*c=='i') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_i);
        } else {
          delchanfla2(tmpcp,cm_i);
        }
      }
      if (*c=='p') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_p);
        } else {
          delchanfla2(tmpcp,cm_p);
        }
      }
      if (*c=='s') {
        if (t2!=1) {
          setchanfla2(tmpcp,cm_s);
        } else {
          delchanfla2(tmpcp,cm_s);
        }
      }
      if (*c=='l') {
        if (t2!=1) {
          if (paramcount<=t1) {
            sprintf(logline,"!!! Failed to parse modes: %s !!!",lastline); putlog(); return;
          }
          setchanli2(tmpcp,atol(params[t1]));
          setchanfla2(tmpcp,cm_l);
          t1++;
        } else {
          delchanfla2(tmpcp,cm_l);
        }
      }
      if (*c=='k') {
        if (paramcount<=t1) {
          sprintf(logline,"!!! Failed to parse modes: %s !!!",lastline); putlog(); return;
        }
        setchanke2(tmpcp,params[t1]);
        if (t2!=1) {
          setchanfla2(tmpcp,cm_k);
        } else {
          delchanfla2(tmpcp,cm_k);
        }
        t1++;
      }
      if (*c=='b') {
        t1++;
        // Real Handling still missing
      }
    }
  }
}

void handleprivmsg() {
  static char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], ucmd[TMPSSIZE], lotmp[TMPSSIZE];
  int res; long unum; static char tail[TMPSSIZE]; int iio;
  res=sscanf(lastline,"%s %s %s %s",tmps2,tmps3,tmps4,ucmd);
  if (res!=4) {
    sprintf(logline,"!!! Failed to parse privmsg: %s !!!",lastline); putlog();
    return;
  }
  unum=tokentolong(sender);
  toLowerCase(ucmd);
  strcpy(tmps3,&ucmd[1]);
  strcpy(ucmd,tmps3);
//  printf("%d :: %s\n",&unum,ucmd);
  res=strcspn(&lastline[1],":");
  strcpy(tail,&lastline[res+2]);
  iio=isircop(unum);
  if (!iio) {
    tail[250]='\0';  // Truncate to max. 250 characters for non-ircops
  } else {
    tail[450]='\0';  // And 450 for opers to prevent too long lines
  }
//  printf("Tail: %s ;\n",tail);
  if (ischarinstr('@',tmps4)) {
    toLowerCase(tmps4);
  } else {
    normnum(tmps4);
    sprintf(tmps2,"%sAAB",servernumeric);
    if (strcmp(tmps4,tmps2)!=0) {
      /* That message was not meant for O, but for one of the fakeusers */
      return;
    }
  }
  numtohostmask(unum,lotmp);
  if ((strcmp(ucmd,"password")!=0) && (strcmp(ucmd,"auth"))!=0) {
    sprintf(logline,"%s !%s!",lotmp,tail); putlog();
  }
  if ((strcmp(tmps4,nicktomsg)!=0) && (strcmp(ucmd,"auth")==0)) {
    msgtouser(unum,"You should not /msg me this way. If you want to send commands, you should use");
    sprintf(tmps4,"/msg %s@%s",mynick,myname);
    msgtouser(unum,tmps4);
  }
// Insert all commands that are available to non-ircops here - and a copy below
  if (strcmp(ucmd,"auth")==0) { doauth(unum,tail); return; }
  if (strcmp(ucmd,"help")==0) { douserhelp(unum,tail); }
  if (strcmp(ucmd,"showcommands")==0) { douserhelp(unum,tail); }
  if (dodyncmds(ucmd,unum,tail,0)) { return; }
  if (iio) {
// Commands only available for ircops
    if (strcmp(ucmd,"hello")==0) { dohello(unum,tail); return; }
    if (strcmp(ucmd,"password")==0) { dochangepwd(unum,tail); return; }
    if (strcmp(ucmd,"noticeme")==0) { donoticeset(unum,tail); return; }
    if (strcmp(ucmd,"whois")==0) { dowhois(unum,tail); return; }
    if (strcmp(ucmd,"status")==0) { dostatus(unum,tail); return; }
    if (strcmp(ucmd,"opchan")==0) { doopchan(unum,tail); return; }
    if (strcmp(ucmd,"jupe")==0) { dojupe(unum,tail); return; }
    if (strcmp(ucmd,"help")==0) { dooperhelp(unum,tail); return; }
    if (strcmp(ucmd,"showcommands")==0) { dooperhelp(unum,tail); return; }
    if (strcmp(ucmd,"channel")==0) { dochancmd(unum,tail); return; }
    if (strcmp(ucmd,"resync")==0) { doresync(unum,tail); return; }
    if (strcmp(ucmd,"semiresync")==0) { dosemiresync(unum,tail); return; }
    if (strcmp(ucmd,"glist")==0) { doglist(unum,tail); return; }
    if (strcmp(ucmd,"gline")==0) { dogline(unum,tail); return; }
    if (strcmp(ucmd,"ungline")==0) { doungline(unum,tail); return; }
    if (strcmp(ucmd,"countusers")==0) { docountusers(unum,tail); return; }
    if (strcmp(ucmd,"mfc")==0) { domfc(unum,tail); return; }
    if (strcmp(ucmd,"settime")==0) { dosettimecmd(unum,tail); return; }
    if (strcmp(ucmd,"serverlist")==0) { doserverlist(unum,tail); return; }
    if (strcmp(ucmd,"changelev")==0) { dochangelev(unum,tail); return; }
    if (strcmp(ucmd,"deluser")==0) { dodelusercmd(unum,tail); return; }
    if (strcmp(ucmd,"kick")==0) { dokickcmd(unum,tail); return; }
    if (strcmp(ucmd,"broadcast")==0) { dobroadcast(unum,tail); return; }
    if (strcmp(ucmd,"syncglines")==0) { dosyncglines(unum,tail); return; }
    if (strcmp(ucmd,"subnetlist")==0) { dosubnetlist(unum,tail); return; }
    if (strcmp(ucmd,"listauthed")==0) { dolistauthed(unum,tail); return; }
    if (strcmp(ucmd,"save")==0) { dosave(unum,tail); return; }
    if (strcmp(ucmd,"sntrustlist")==0) { dosntrustlist(unum,tail); return; }
    if (strcmp(ucmd,"sntrustadd")==0) { dosntrustadd(unum,tail); return; }
    if (strcmp(ucmd,"sntrustdel")==0) { dosntrustdel(unum,tail); return; }
    if (strcmp(ucmd,"sbroadcast")==0) { dosbroadcast(unum,tail); return; }
    if (strcmp(ucmd,"deopall")==0) { dodeopall(unum,tail); return; }
    if (strcmp(ucmd,"spew")==0) { dospew(unum,tail); return; }
    if (strcmp(ucmd,"fakeuser")==0) { docreatefakeuser(unum,tail); return; }
    if (strcmp(ucmd,"fakeadd")==0) { docreatefakeuser(unum,tail); return; }
    if (strcmp(ucmd,"rnc")==0) { dornc(unum,tail); return; }
    if (strcmp(ucmd,"rngline")==0) { dorngline(unum,tail); return; }
    if (strcmp(ucmd,"rnungline")==0) { dornungline(unum,tail); return; }
    if (strcmp(ucmd,"rnglist")==0) { dornglist(unum,tail); return; }
    if (strcmp(ucmd,"fakelist")==0) { dofakelist(unum,tail); return; }
    if (strcmp(ucmd,"fakekill")==0) { dofakekill(unum,tail); return; }
    if (strcmp(ucmd,"fakedel")==0) { dofakekill(unum,tail); return; }
    if (strcmp(ucmd,"inttobase64")==0) { dointtobase64(unum,tail); return; }
    if (strcmp(ucmd,"base64toint")==0) { dobase64toint(unum,tail); return; }
    if (strcmp(ucmd,"splitlist")==0) { dosplitlist(unum, tail); return; }
    if (strcmp(ucmd,"splitdel")==0) { dosplitdel(unum, tail); return; }
    if (strcmp(ucmd,"loadmod")==0) { doloadmod(unum, tail); return; }
    if (strcmp(ucmd,"unloadmod")==0) { dounloadmod(unum, tail); return; }
    if (strcmp(ucmd,"reloadmod")==0) { doreloadmod(unum, tail); return; }
    if (strcmp(ucmd,"lsmod")==0) { dolsmod(unum, tail); return; }
    if (strcmp(ucmd,"die")==0) { dodie(unum, tail); return; }
    if (strcmp(ucmd,"noticemask")==0) { donoticemask(unum, tail); return; }
    if (strcmp(ucmd,"trustgroupadd")==0) { dotrustgroupadd(unum, tail); return; }
    if (strcmp(ucmd,"trustgroupdel")==0) { dotrustgroupdel(unum, tail); return; }
    if (strcmp(ucmd,"trustadd")==0) { dotrustadd(unum, tail); return; }
    if (strcmp(ucmd,"trustdel")==0) { dotrustdel(unum, tail); return; }
    if (strcmp(ucmd,"trustlist")==0) { dotrustlist(unum, tail); return; }
    if (strcmp(ucmd,"trustgline")==0) { dotrustgline(unum, tail); return; }
    if (strcmp(ucmd,"trustspew")==0) { dotrustspew(unum, tail); return; }
    if (dodyncmds(ucmd,unum,tail,1)) { return; }
    msgtouser(unum,"Unknown command.");
  } else {
    msgtouser(unum,"This command is either unknown, or you need to be opered up to use it.");
  }
}

void handleburstmsg() {
  channel *ctmp; long creationtime; int actflags; char *c; long bibabutzelmann;
  int xofs; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; int i; int k;
  if (paramcount<5) {
    sprintf(logline,"!!! Failed to parse Burst-Message: %s !!!",lastline); putlog();
    return;
  }
  creationtime=strtol(params[3],NULL,10);
  toLowerCase(params[2]);
  ctmp=getchanptr(params[2]);
  if (ctmp==NULL) {
    if (strlen(params[2])>CHANNELLEN) {
      sprintf(logline,"!!! Channelname exceeds CHANNELLEN: %s !!!",lastline); putlog();
      return;
    }
    ctmp=newchan(params[2],creationtime);
  }
  if (params[4][0]=='+') {
    // that channel has some modes (like +t etc.)
    int j;
    xofs=5;
    c=&params[4][1];
    for (;*c!='\0';c++) {
      if (*c=='k') {
        // Channel has a key
        if (paramcount<=xofs) {
          sprintf(logline,"!!! Failed to parse BURST (missing arg to chanflag): %s !!!",lastline);
          putlog(); return;
        }
        setchanke2(ctmp,params[xofs]);
        xofs++;
      }
      if (*c=='l') {
        // Channel has a limit
        if (paramcount<=xofs) {
          sprintf(logline,"!!! Failed to parse BURST (missing arg to chanflag): %s !!!",lastline);
          putlog(); return;
        }
        setchanli2(ctmp,strtol(params[xofs],NULL,10));
        xofs++;
      }
      for (j=0;j<numcf;j++) {
        if (chanflags[j].c==*c) { setchanfla2(ctmp,chanflags[j].v); }
      }
    }
  } else {
    xofs=4;
  }
  // modes have been handled if they existed. params[xofs] is the userlist.
  if (paramcount<=xofs) { 
    sprintf(logline,"IRC-Bug: %s",lastline); putlog();
    return; /* Means we got a channel without any users - ircu-bug */
  } 
  actflags=0;
  if (strcmp(lastburstchan,params[2])==0) { actflags=lastburstflags; }
  c=&params[xofs][0];
  if (*c!=':') { /* if it begins with :% it's just a banlist! */
    for (;(*c!='\0');c++) {
      i=0;
      for (;((*c!='\0') && (*c!=','));c++) {
        tmps2[i]=*c; i++;
      }
      tmps2[i]='\0';
      if (strlen(tmps2)<3) { break; }
      if (strlen(tmps2)==3) { // V.07 numeric, no modes
        normnum(tmps2); tmps3[0]='\0';
      } else {
        if (tmps2[3]==':') { // V.07 Numeric with modes
          strcpy(tmps3,&tmps2[4]);
          tmps2[3]='\0';
          normnum(tmps2);
        } else {
          if (tmps2[5]==':') {
            strcpy(tmps3,&tmps2[6]);
            tmps2[5]='\0';
          } else {
            tmps3[0]='\0';
          }
        }
      }
      /* tmps2 now contains the normed numeric, tmps3 the new modes if any. */
      if (tmps3[0]!='\0') { actflags=0; }
      for (i=0;i<strlen(tmps3);i++) {
        for (k=0;k<numuf;k++) {
          if (userflags[k].c==tmps3[i]) { actflags=actflags | userflags[k].v; }
        }
      }
      bibabutzelmann=tokentolong(tmps2);
      addusertocptr(ctmp,bibabutzelmann);
      changechanmod2(ctmp,bibabutzelmann,0,actflags);
      addchantouser2(bibabutzelmann,ctmp);
      if (*c=='\0') { break; }
    }
  }
  strcpy(lastburstchan,params[2]);
  lastburstflags=actflags;
}

void handlecreatemsg() {
/* [15:28] DAO CREATE #fook,#foo2 971097981 */
  long creattime; char *c; int i; char tmps2[TMPSSIZE]; long num;
  channel *ctmp;
  if (paramcount<4) {
    sprintf(logline,"!!! Error parsing create message: %s !!!",lastline); putlog();
    return;
  }
  toLowerCase(params[2]);
  creattime=strtol(params[3],NULL,10);
  num=tokentolong(sender);
  c=&params[2][0];
  for (;(*c!='\0');c++) {
    i=0;
    for (;((*c!='\0') && (*c!=','));c++) {
      tmps2[i]=*c; i++;
    }
    tmps2[i]='\0';
//    printf("DEBUG: Creating chan %s\n",tmps2);
    if (chanexists(tmps2)!=1) { newchan(tmps2,creattime); }
    addusertochan(tmps2,num);
    changechanmode(tmps2,num,0,um_o);
    ctmp=getchanptr(tmps2);
    if (tmps2!=NULL) { addchantouser2(num,ctmp); }
    if (*c=='\0') { break; }
  }
}

void handlejoinmsg() {
  long num; char *c; int i; char tmps2[TMPSSIZE]; channel *ctmp;
  if (paramcount<3) {
    sprintf(logline,"!!! Error parsing join message: %s !!!",lastline); putlog();
    return;
  }
  num=tokentolong(sender);
  toLowerCase(params[2]);
  c=&params[2][0];
  for (;(*c!='\0');c++) {
    i=0;
    for (;((*c!='\0') && (*c!=','));c++) {
      tmps2[i]=*c; i++;
    }
    tmps2[i]='\0';
    if (strcmp(tmps2,"0")==0) {
      deluserfromallchans(num);
    } else {
      if (!chanexists(tmps2)) {
        sprintf(logline,"!!! JOIN for nonexistant channel: %s !!!",lastline); putlog();
        newchan(tmps2,0);
      }
      addusertochan(tmps2,num);
      ctmp=getchanptr(tmps2);
      if (tmps2!=NULL) { addchantouser2(num,ctmp); }
    }
    if (*c=='\0') { break; }
  }
}

void handlepartmsg() {
  long num; char *parspos;
  if (paramcount<3) {
    sprintf(logline,"!!! PART failed: too few parameters: %s !!!",lastline); putlog();
    return;
  }
  num=tokentolong(sender);
  toLowerCase(params[2]);
  parspos=strtok(params[2],",");
  while (parspos!=NULL) {
    delchanfromuser(num,parspos);
    deluserfromchan(parspos,num);
    parspos=strtok(NULL,",");
  }
}

void handlekickmsg() {
/* [01:11] :Fox_Muld_ KICK #twilightzone DAL :That's gotta hurt! */
  long num; char *c; int i; char tmps2[TMPSSIZE];
  if (paramcount<4) {
    sprintf(logline,"!!! Error parsing kick message: %s !!!",lastline); putlog();
    return;
  }
  toLowerCase(params[2]);
  c=&params[3][0];
  for (;(*c!='\0');c++) {
    i=0;
    for (;((*c!='\0') && (*c!=','));c++) {
      tmps2[i]=*c; i++;
    }
    tmps2[i]='\0';
    normnum(tmps2);
    num=tokentolong(tmps2);
    delchanfromuser(num,params[2]);
    deluserfromchan(params[2],num);
    if (*c=='\0') { break; }
  }
}

void handleglinemsg() {
  int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE], tmps7[TMPSSIZE];
  res=sscanf(lastline,"%s %s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6,tmps7);
  if (res<4) {
    sprintf(logline,"!!! Error parsing gline message: %s !!!",lastline); putlog();
    return;
  }
  if (tmps5[0]=='+') {
    if (res!=6) {
      sprintf(logline,"!!! Failed to add gline from foreign server (Parameters missing): %s !!!",lastline); putlog();
      return;
    }
    addgline(&tmps5[1],&tmps7[1],"Network",atol(tmps6),0);
  } else {
    remgline(&tmps5[1],0);
  }
}

void handlestatsPreply() {
/* [00:00] :leaf.de.quakenet.eu.org 217 Operserv2 P 6668 0 :0x2000
   [00:00] :leaf.de.quakenet.eu.org 217 Operserv2 P 6667 2 :0x2000 */
  int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE];
  long servn; serverdata *s;
  res=sscanf(lastline,"%s %s %s %s %s",tmps2,tmps3,tmps4,tmps5,tmps6);
  if (res!=5) {
    sprintf(logline,"!!! Failed to parse /stats P - reply: %s !!!", lastline); putlog();
    return;
  }
  if (tmps2[0]==':') {
    servn=getsrvnum(&tmps2[1]);
  } else {
    servn=tokentolong(tmps2);
  }
  s=sls;
  while (s!=NULL) {
    if (s->numeric==servn) {
      // this is the server we got the stats-reply for
      if (strlen(s->ports)>1000) { return; }
      if (strlen(s->ports)==0) {
        strcpy(s->ports, tmps6);
      } else {
        strcat(s->ports, ",");
        strcat(s->ports, tmps6);
      }
      return;
    }
    s=(void *)s->next;
  }
  sprintf(logline,"!!! Got /stats P - reply from unknown server! !!!"); putlog();
}

void handlestatsureply() {
/* [00:32] :leaf.de.quakenet.eu.org 242 Operserv2 :Server Up 65 days, 23:48:14 */
  int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE], tmps7[TMPSSIZE];
  char *parspos; long secsup; long servn; serverdata *s;
  res = sscanf(lastline,"%s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5);
  if (res!=4) {
    sprintf(logline,"!!! Failed to parse /stats u - reply(#1): %s !!!",lastline); putlog();
    return;
  }
  res=sscanf(tmps5,"%s %s %s %s %[^\n]",tmps6,tmps6,tmps3,tmps6,tmps4);
  if (res!=5) {
    sprintf(logline,"!!! Failed to parse /stats u - reply(#2): %s !!!",lastline); putlog();
    return;
  }
  parspos=strtok(tmps4,":"); strcpy(tmps5,parspos);
  parspos=strtok(NULL,":"); strcpy(tmps6,parspos);
  parspos=strtok(NULL,":"); strcpy(tmps7,parspos);
//  printf("Days: %s ; Rest: %s %s %s\n",tmps3,tmps5,tmps6,tmps7);
  secsup=atol(tmps7)+atol(tmps6)*60+atol(tmps5)*3600+atol(tmps3)*86400;
  if (tmps2[0]==':') {
    servn=getsrvnum(&tmps2[1]);
  } else {
    servn=tokentolong(tmps2);
  }
  s=sls;
  while (s!=NULL) {
    if (s->numeric==servn) {
      s->createdat=getnettime()-secsup;
//      printf("%d up for %d secs\n",s->numeric,secsup);
      s->pingreply=getnettime();
      if (gotallpingreplys()) {
        if (getmaxping()>5) {
          sprintf(logline,"Network has more than 5 seconds (%ld) lag, cannot do settime!",getmaxping()); putlog();
          lastsettime+=5*60; waitingforping=0;
        } else {
          sprintf(logline,"Network lag: %ld seconds. Sending settime.",getmaxping()); putlog();
          timestdif=0;  /* _WE_ are authoritative with respect to time. */
          lastsettime=getnettime(); waitingforping=0;
          fprintf(sockout,"%s SE %lu\n",servernumeric,(long int)getnettime());
          fflush(sockout);
        }
      }
      return;
    }
    s=(void *)s->next;
  }
  sprintf(logline,"!!! Got /stats u - reply from unknown server: %s !!!",lastline); putlog();
}

void handlestatsmsg() {
  int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE]; long num;
  res=sscanf(lastline,"%s %s %s",tmps2,tmps3,tmps4);
  if (res<3) { return; }
  if (tmps2[0]==':') {
    toLowerCase(tmps2);
    num=nicktonum(&tmps2[1]);
  } else {
    normnum(tmps2);
    num=tokentolong(tmps2);
  }
  longtotoken(num,tmps3,5);
  if (tmps3[0]=='\0') { return; }
  if (strcmp(tmps4,"P")==0) {
    /* /stats P requested (open ports) */
    fprintf(sockout,"%s 217 %s P none 0 0x2000\r\n",servernumeric,tmps3);
    fprintf(sockout,"%s 219 %s P :End of /STATS report\r\n",servernumeric,tmps3);
    fflush(sockout);
  }
  if (strcmp(tmps4,"u")==0) {
    /* /stats u requested (uptime) */
    long iamup, days, hours, minutes;
    iamup=getnettime()-starttime;
    days=iamup/(24*3600); iamup%=(24*3600);
    hours=iamup/3600; iamup%=3600;
    minutes=iamup/60; iamup%=60;
    sprintf(tmps2,"%ld days, %02ld:%02ld:%02ld",days,hours,minutes,iamup);
    fprintf(sockout,"%s 242 %s :Server Up %s\r\n",servernumeric,tmps3, tmps2);
    fprintf(sockout,"%s 250 %s :Highest connection count: 2 (1 client)\r\n",servernumeric,tmps3);
    fprintf(sockout,"%s 219 %s u :End of /STATS report\r\n",servernumeric,tmps3);
    fflush(sockout);
  }
}

void handleeoback() {
  serverdata *a; char tmpn[SRVLEN+1]; long num;
  num=tokentolong(sender);
  a=sls;
  while (a!=NULL) {
    if (num==a->numeric) {
      strncpy(tmpn,a->nick,SRVLEN);
      tmpn[SRVLEN]='\0';
      toLowerCase(tmpn);
      delsplit(tmpn);
      return;
    }
    a=(void *)a->next;
  }
  sprintf(logline,"uups, EOB_ACK from unknown server?!"); putlog();
}

void handletopic() {
  channel *c; char tmps[TMPSSIZE]; char *a, *b;
  c=getchanptr(params[2]);
  if (c==NULL) { return; }
  a=&lastline[0]; b=&tmps[0];
  while ((*a!=':') && (*a!='\0')) { a++; }
  if (*a=='\0') { strcpy(c->topic,""); return; }
  a++;
  while (*a!='\0') {
    *b=*a; a++; b++;
  }
  *b='\0';
  strncpy(c->topic,tmps,TOPICLEN);
  c->topic[TOPICLEN]='\0';
}
