/*
Operservice 2
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "mysql_version.h"
#include "mysql.h"
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "globals.h"

/* subnetlist-trust functions */

void createtl() {
  tsn=(trustedsubnet *) malloc(tsnsi*sizeof(trustedsubnet));
  if (tsn==NULL) {
    printf("Memory allocation failed.\n");
    exit(0);
  }
}

void snltrustadd(unsigned int a, unsigned int b, unsigned int c, unsigned int n,
                 unsigned long expires, char *reason) {
  int i; trustedsubnet *t1;
  if (tsnsi<=tsncu+1) {
    /* List is full, we need to make it bigger */
    tsnsi+=100;
    tsn=(trustedsubnet *)realloc((void *) tsn, tsnsi*sizeof(trustedsubnet));
    if (tsn==NULL) {
      sprintf(logline,"!!! MEMORY REALLOCATION FOR TRUSTLIST FAILED! !!!\n"); putlog();
      exit(-10);
    }
  }
  t1=tsn;
  for (i=0;i<tsncu;i++) {
    if ((t1->a==a) && (t1->b==b) && (t1->c==c)) {
      t1->n=n; t1->expires=expires; strcpy(t1->reason,reason); return;
    }
    t1++;
  }
  t1->a=a; t1->b=b; t1->c=c; t1->n=n; t1->expires=expires;
  strcpy(t1->reason,reason); tsncu++;
}

int snltrustdel(unsigned int a, unsigned int b, unsigned int c) {
  int i; trustedsubnet *t1;
  t1=tsn;
  for (i=0;i<tsncu;i++) {
    if ((t1->a==a) && (t1->b==b) && (t1->c==c)) {
      /* Kill that subnet from the trustlist */
      tsncu--;
      if (tsncu!=i) { 
        memmove(t1,t1+1,((tsncu-i)*sizeof(trustedsubnet)));
      }
      return 1;
    }
    t1++;
  }
  return 0;
}

unsigned int getsnltrust(unsigned long IP) {
  int i; trustedsubnet *t1; unsigned int a,b,c,d;
  splitip(IP,&a,&b,&c,&d);
  t1=tsn;
  for (i=0;i<tsncu;i++) {
    if ((t1->a==a) && (t1->b==b) && (t1->c==c)) {
      if (t1->expires>getnettime()) {
        return t1->n;
      } else {
        trustlistexpire();
        return mfwarn;
      }
    }
    t1++;
  }
  return mfwarn;
}

void trustlistexpire() {
  int i; trustedsubnet *t1;
  t1=tsn;
  for (i=0;i<tsncu;i++) {
    if (t1->expires<getnettime()) {
      /* Kill that subnet from the trustlist */
      tsncu--;
      if (tsncu!=i) { memmove(t1,t1+1,((tsncu-i)*sizeof(trustedsubnet))); }
      i--;
    }
    t1++;
  }
}

void loadtrustlist() {
  int res;
  MYSQL_RES *myres; MYSQL_ROW myrow;
  res=mysql_query(&sqlcon,"SELECT * FROM sntrustlist");
  if (res!=0) {
    sprintf(logline,"!!! Failed to query subnet-trust-list from database !!!"); putlog();
    return;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=6) {
    sprintf(logline,"!!! TrustList in database has invalid format !!!"); putlog();
    return;
  }
  tsncu=0;
  while ((myrow=mysql_fetch_row(myres))) {
    if (strlen(myrow[5])>200) {
      sprintf(logline,"!!! Malformed trustlist-entry in database! !!!"); putlog();
    } else {
      snltrustadd(atoi(myrow[0]),atoi(myrow[1]),atoi(myrow[2]),atoi(myrow[3]),strtoul(myrow[4],NULL,10),myrow[5]);
    }
  }
  mysql_free_result(myres);
  trustlistexpire();
}

void savetrustlist() {
  char query[TMPSSIZE]; int i; trustedsubnet *t1; char tmps2[TMPSSIZE];
  trustlistexpire();
  mysql_query(&sqlcon,"DROP TABLE IF EXISTS sntrustlist");
  mysql_query(&sqlcon,"CREATE TABLE sntrustlist (a INT, b INT, c INT, n INT, expires BIGINT, reason VARCHAR(200))");
  t1=tsn;
  for (i=0;i<tsncu;i++) {
    mysql_escape_string(tmps2,t1->reason,strlen(t1->reason));
    sprintf(query,"INSERT INTO sntrustlist VALUES(%d,%d,%d,%d,%ld,'%s')",t1->a,t1->b,t1->c,t1->n,t1->expires,tmps2);
    mysql_query(&sqlcon,query);
    t1++;
  }
}

/* End of subnetlist-trust-functions */
