/*
Operservice 2 - trusts.c
Manages Trusts for hosts or groups of hosts
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "mysql_version.h"
#include "mysql.h"
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "globals.h"

unsigned int hlhash(char *host) {
  return (unsigned int)(crc32(host)%SIZEOFHL);
}

unsigned int tlhash(char *host) {
  return (unsigned int)(crc32(host)%SIZEOFTL);
}

unsigned int tgshash(unsigned long ID) {
  return (unsigned int)(ID%SIZEOFIDMAP);
}

/* Returns 0 if host isn't trusted, or the ID of the trusted host group */
unsigned long istrusted(char *host) {
  trustedhost *a;
  a=trustedhosts[tlhash(host)];
  while (a!=NULL) {
    if (strcmp(host,a->hostname)==0) { return a->id; }
    a=(void *)a->next;
  }
  return 0;
}

int addtountrustedhost(char *host) {
  unsigned int n; unsigned long i; hostcounter *a;
  n=hlhash(host);
  a=(hostcounter *)(untrustedhosts[n].content);
  for (i=0;i<untrustedhosts[n].cursi;i++) {
    if (strcmp(host,a[i].hostname)==0) {
      a[i].currenton++;
      return a[i].currenton;
    }
  }
  i=array_getfreeslot(&(untrustedhosts[n]));
  a=(hostcounter *)(untrustedhosts[n].content);
  strcpy(a[i].hostname,host);
  a[i].currenton=1;
  return 1;
}

void delfromuntrustedhost(char *host) {
  unsigned int n; unsigned long i; hostcounter *a;
  n=hlhash(host);
  a=(hostcounter *)(untrustedhosts[n].content);
  for (i=0;i<untrustedhosts[n].cursi;i++) {
    if (strcmp(host,a[i].hostname)==0) {
      a[i].currenton--;
      if (a[i].currenton==0) {
        array_delslot(&(untrustedhosts[n]),i);
      }
      return;
    }
  }
}

void addtotrustedhost(char *ident, unsigned long id) {
  trustedgroup *tg; char tmps2[TMPSSIZE]; unsigned long i; identcounter *ic;
  tg=findtrustgroupbyID(id);
  if (tg==NULL) { return; }
  tg->currentlyon++;
  if (tg->currentlyon > tg->trustedfor) {
    sprintf(tmps2,"Trustgroup %s exceeding trustlimit - %d / %d",tg->name,tg->currentlyon,tg->trustedfor);
    sprintf(logline,"[trustexceeded] %s",tmps2); putlog();
    noticeallircops(tmps2);
  }
  if (strlen(ident)>USERLEN) { return; }
  ic=(identcounter *)(tg->identcounts.content);
  for (i=0;i<tg->identcounts.cursi;i++) {
    if (strcmp(ic[i].ident,ident)==0) {
      ic[i].currenton++;
      if ((ic[i].currenton > tg->maxperident) && (tg->maxperident!=0)) {
        sprintf(tmps2,"User %s @ Group %s has excessive connections (%d / %d)",ident,tg->name,ic[i].currenton,tg->maxperident);
        sprintf(logline,"[excon] %s",tmps2); putlog();
        noticeallircops(tmps2);
      }
      return;
    }
  }
  i=array_getfreeslot(&(tg->identcounts));
  ic=(identcounter *)(tg->identcounts.content);
  ic[i].currenton=1;
  strcpy(ic[i].ident,ident);
}

void delfromtrustedhost(char *ident, unsigned long id) {
  trustedgroup *tg; unsigned long i; identcounter *ic;
  tg=findtrustgroupbyID(id);
  if (tg==NULL) { return; }
  tg->currentlyon--;
  if (strlen(ident)>USERLEN) { return; }
  ic=(identcounter *)(tg->identcounts.content);
  for (i=0;i<tg->identcounts.cursi;i++) {
    if (strcmp(ic[i].ident,ident)==0) {
      ic[i].currenton--;
      if (ic[i].currenton==0) {
        array_delslot(&(tg->identcounts),i);
      }
      return;
    }
  }
}

void trustnewclient(char *ident, char *host) {
  char tmph[HOSTLEN+1]; unsigned long thid; char tmps2[TMPSSIZE];
  if (strlen(ident)>USERLEN) { return; }
  if (strlen(host)>HOSTLEN) { return; }
  strcpy(tmph,host); toLowerCase(tmph);
  /* First check if this host is trusted */
  thid=istrusted(tmph);
  if (thid!=0) { /* That host is trusted */
    addtotrustedhost(ident, thid);
  } else { /* untrusted host */
    int currenton;
    currenton=addtountrustedhost(tmph);
    if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
      if (defaultclonewarn==0) { return; }
      sprintf(tmps2,"[%d] clones detected from %s",currenton,host);
      noticeallircops(tmps2);
      sprintf(logline,"[clones] %d clones detected from untrusted host %s",currenton,host);
      putlog();
    }
  }
}

void trustdelclient(char *ident, char *host) {
  char tmph[HOSTLEN+1]; unsigned long thid;
  strcpy(tmph,host); toLowerCase(tmph);
  /* First check if this host is trusted */
  thid=istrusted(tmph);
  if (thid!=0) { /* That host is trusted */
    delfromtrustedhost(ident, thid);
  } else { /* untrusted host */
    delfromuntrustedhost(tmph);
  }
}

trustedgroup * findtrustgroupbyID(unsigned long ID) {
  trustedgroup *tmp;
  tmp=trustedgroups[tgshash(ID)];
  while (tmp!=NULL) {
    if (ID==tmp->id) {
      return tmp;
    }
    tmp=(void *)tmp->next;
  }
  return NULL;
}

trustedgroup * findtrustgroupbyname(char *name) {
  trustedgroup *tmp; int i;
  for (i=0;i<SIZEOFIDMAP;i++) {
    tmp=trustedgroups[i];
    while (tmp!=NULL) {
      if (strcmp(name,tmp->name)==0) {
        return tmp;
      }
      tmp=(void *)tmp->next;
    }
  }
  return NULL;
}

unsigned long getfreetgid() {
  unsigned long id; trustedgroup *tg;
  id=0;
  do {
    id++;
    tg=findtrustgroupbyID(id);
  } while (tg!=NULL);
  return id;
}

void dotrustgroupadd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  char tmps6[TMPSSIZE], tmps7[TMPSSIZE], tmps8[TMPSSIZE], tmps9[TMPSSIZE];
  long dur; trustedgroup *tg;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6,tmps7,tmps8,tmps9);
  if ((res<7) || (res>8)) {
    msgtouser(unum,"Syntax: trustgroupadd name howmany howlong maxperident enforceident contact [comment]");
    msgtouser(unum,"where name is a unique name for that group");
    msgtouser(unum,"      howmany gives the number of clones allowed from that group");
    msgtouser(unum,"      howlong gives how long the trust will last (e.g. '1y')");
    msgtouser(unum,"      maxperident sets how many clones are allowed from a single user.");
    msgtouser(unum,"                  0 means unlimited");
    msgtouser(unum,"      enforceident if set to 1 will kill all clients without idents");
    msgtouser(unum,"                   in that group");
    msgtouser(unum,"      contact is one or more contact-emails, seperated by commas,");
    msgtouser(unum,"              WITH NO SPACES!!!");
    msgtouser(unum,"      comment is an optional comment");
    return;
  }
  if (res==7) { strcpy(tmps9,"[no comment]"); }
  if (strlen(tmps3)>TRUSTNAMELEN) {
    msgtouser(unum,"The name you gave is too long");
    return;
  }
  if (tmps3[0]=='#') {
    msgtouser(unum,"Trustgroupname must not start with a '#'");
    return;
  }
  if (strlen(tmps8)>TRUSTCONTACTLEN) {
    msgtouser(unum,"contact-email is too long"); return;
  }
  if (strlen(tmps9)>TRUSTCOMLEN) {
    msgtouser(unum,"comment is too long"); return;
  }
  dur=durationtolong(tmps5);
  if (dur<1) {
    msgtouser(unum,"Invalid duration given");
    return;
  }
  if (((tmps7[0]!='0') && (tmps7[0]!='1')) || (tmps7[1]!='\0')) {
    msgtouser(unum,"enforceident is a boolean setting, that means it can only be 0 or 1"); return;
  }
  toLowerCase(tmps3);
  tg=findtrustgroupbyname(tmps3);
  if (tg!=NULL) { /* This group exists, modify it */
    sprintf(tmps2,"Trustgroup %s already exists, replacing values.",tmps3);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Old settings were: trusted for %d users, max. clones per user %d, %senforcing ident",
            tg->trustedfor,tg->maxperident,(tg->enforceident==0) ? "not " : "");
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Contact-eMail: %s",tg->contact);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Comment:       %s",tg->comment);
    msgtouser(unum,tmps2);
  } else { /* New group */
    unsigned long newID;
    tg=(trustedgroup *)malloc(sizeof(trustedgroup));
    if (tg==NULL) {
      printf("!!! Out of memory in dotrustgroupadd !!!\n"); exit(0);
    }
    newID=getfreetgid();
    tg->id=newID;
    tg->currentlyon=0;
    tg->next=(void *)trustedgroups[tgshash(newID)];
    trustedgroups[tgshash(newID)]=tg;
    array_init(&(tg->identcounts),sizeof(identcounter));
  }
  tg->enforceident=(tmps7[0]=='1') ? 1 : 0;
  strcpy(tg->name,tmps3);
  strcpy(tg->contact,tmps8);
  strcpy(tg->comment,tmps9);
  tg->trustedfor=strtol(tmps4,NULL,10);
  tg->maxperident=strtol(tmps6,NULL,10);
  tg->expires=getnettime()+dur;
  msgtouser(unum,"Trustgroup created/updated.");
}

int destroytrustgroup(unsigned long id) {
  array listofhosts; trustedhost *th; long i, j; trustedhost *th2, *th3;
  trustedgroup *tg1, *tg2, *tg3; userdata *u; char tmph[HOSTLEN+1]; char tmps2[TMPSSIZE];
  /* Not only will we have to delete the trustgroup, we will also have to recount
     all individual hosts in it. So lets start with building a list of all hosts */
  array_init(&listofhosts,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    th2=NULL;
    while (th!=NULL) {
      if (th->id==id) {
        /* This host belongs to our trustgroup */
        long tmpl=array_getfreeslot(&listofhosts);
        strcpy(listofhosts.content+tmpl*(HOSTLEN+1),th->hostname);
        if (th2==NULL) {
          trustedhosts[i]=(void *)th->next;
        } else {
          th2->next=th->next;
        }
        th3=th;
        th=(void *)th->next;
        free(th3);
      } else {
        th2=th;
        th=(void *)th->next;
      }
    }
  }
  /* we now have the list in array listofhosts, and the hosts deleted from trustedhosts */
  /* Goodbye trustgroup */
  tg1=trustedgroups[tgshash(id)]; tg2=NULL;
  while (tg1!=NULL) {
    if (tg1->id==id) { /* Kill it! */
      array_free(&(tg1->identcounts));
      if (tg2==NULL) {
        trustedgroups[tgshash(id)]=(void *)tg1->next;
      } else {
        tg2->next=tg1->next;
      }
      tg3=tg1;
      tg1=(void *)tg1->next;
      free(tg3);
    } else {
      tg2=tg1;
      tg1=(void *)tg1->next;
    }
  }
  /* Recount all hosts - this is gonna be a CPU Killer, but it's used less than
     once a day, so i don't care too much. */
  for (i=0;i<SIZEOFUL;i++) {
    u=uls[i];
    while (u!=NULL) {
      strcpy(tmph,u->host); toLowerCase(tmph);
      for (j=0;j<listofhosts.cursi;j++) {
        if (strcmp(tmph,listofhosts.content+j*(HOSTLEN+1))==0) {
          int currenton=addtountrustedhost(tmph);
          if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
            if (defaultclonewarn!=0) {
              sprintf(tmps2,"[%d] clones detected from %s",currenton,u->host);
              noticeallircops(tmps2);
              sprintf(logline,"[clones] %d clones detected from untrusted host %s",currenton,u->host);
              putlog();
            }
          }
        }
      }
      u=(void *)u->next;
    }
  }
  i=listofhosts.cursi;
  array_free(&listofhosts);
  return i;
}

void dotrustgroupdel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  trustedgroup *tg;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: trustgroupdel <groupname OR #groupid>");
    msgtouser(unum,"This deletes the group including all hosts so be careful with this!");
    return;
  }
  toLowerCase(tmps3);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  if (tg->id==0) { msgtouser(unum,"Internal error: Trustgroup has ID 0"); return; }
  sprintf(tmps2,"Trustgroup with %d hosts deleted.",destroytrustgroup(tg->id));
  msgtouser(unum,tmps2);
}

void dotrustadd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  trustedgroup *tg; trustedhost *th; long i; userdata *u; char tmph[HOSTLEN+1];
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s",tmps2,tmps3,tmps4,tmps5);
  if (res!=3) {
    msgtouser(unum,"Syntax: trustadd <groupname OR #groupid> host"); return;
  }
  toLowerCase(tmps3);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  if (tg->id==0) { msgtouser(unum,"Internal error: Trustgroup has ID 0"); return; }
  toLowerCase(tmps4);
  if (strlen(tmps4)>HOSTLEN) {
    msgtouser(unum,"Hostname too long"); return;
  }
  if (istrusted(tmps4)>0) {
    msgtouser(unum,"That host already belongs to some trustgroup"); return;
  }
  /* OK so we have a valid group and host - create new hostentry */
  th=(trustedhost *)malloc(sizeof(trustedhost));
  strcpy(th->hostname,tmps4);
  th->id=tg->id;
  th->next=(void *)trustedhosts[tlhash(tmps4)];
  trustedhosts[tlhash(tmps4)]=th;
  /* Remove host from untrusted hosts, and instead add it to the trustgroup! */
  for (i=0;i<SIZEOFUL;i++) {
    u=uls[i];
    while (u!=NULL) {
      strcpy(tmph,u->host); toLowerCase(tmph);
      if (strcmp(tmph,tmps4)==0) {
        delfromuntrustedhost(tmps4);
        addtotrustedhost(u->ident,tg->id);
      }
      u=(void *)u->next;
    }
  }
  sprintf(tmps2,"Added host %s to trustgroup %s (#%lu)",tmps4,tg->name,tg->id);
  msgtouser(unum,tmps2);
}

void dotrustdel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  trustedgroup *tg; trustedhost *th, *t2;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: trustdel host"); return;
  }
  toLowerCase(tmps3);
  th=trustedhosts[tlhash(tmps3)]; t2=NULL;
  while (th!=NULL) {
    if (strcmp(th->hostname,tmps3)==0) {
      long i; userdata *u; char tmph[HOSTLEN+1];
      tg=findtrustgroupbyID(th->id);
      if (t2==NULL) {
        trustedhosts[tlhash(tmps3)]=(void *)th->next;
      } else {
        t2->next=th->next;
      }
      free(th);
      /* Recount the host */
      for (i=0;i<SIZEOFUL;i++) {
        u=uls[i];
        while (u!=NULL) {
          strcpy(tmph,u->host); toLowerCase(tmph);
          delfromtrustedhost(u->ident,tg->id);
          if (strcmp(tmph,tmps3)==0) {
            int currenton=addtountrustedhost(tmph);
            if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
              if (defaultclonewarn!=0) {
                sprintf(tmps2,"[%d] clones detected from %s",currenton,u->host);
                noticeallircops(tmps2);
                sprintf(logline,"[clones] %d clones detected from untrusted host %s",currenton,u->host);
                putlog();
              }
            }
          }
          u=(void *)u->next;
        }
      }
      sprintf(tmps2,"Host removed from Trustgroup %s (#%lu)",tg->name,tg->id);
      msgtouser(unum,tmps2);
      return;
    } else {
      t2=th;
      th=(void *)th->next;
    }
  }
  msgtouser(unum,"That host was not trusted.");
}

/* Trusts have to be loaded BEFORE linking to the net.
   this function returns the number of hosts loaded from the database. */
unsigned long loadtrusts() {
  int res; MYSQL_RES *myres; MYSQL_ROW myrow; unsigned long hostcounter=0;
  res=mysql_query(&sqlcon,"SELECT * FROM trustgroups");
  if (res!=0) {
    sprintf(logline,"!!! Failed to query trustgroups from database !!!"); putlog();
    return 0;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=8) {
    sprintf(logline,"!!! trustgroups in database have invalid format !!!"); putlog();
    return 0;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    if ((strlen(myrow[1])>TRUSTNAMELEN) || (strlen(myrow[6])>TRUSTCONTACTLEN) || (strlen(myrow[7])>TRUSTCOMLEN)) {
      sprintf(logline,"!!! Malformed trustgroups-entry in database! !!!"); putlog();
    } else {
      if (strtol(myrow[0],NULL,10)==0) {
        sprintf(logline,"!!! Trustgroups-ID of 0 in database! !!!"); putlog();
      } else {
        trustedgroup *tg; unsigned int thehash;
        tg=malloc(sizeof(trustedgroup));
        tg->id=strtol(myrow[0],NULL,10);
        strcpy(tg->name,myrow[1]);
        tg->trustedfor=strtol(myrow[2],NULL,10);
        tg->expires=strtol(myrow[3],NULL,10);
        tg->maxperident=strtol(myrow[4],NULL,10);
        tg->enforceident=strtol(myrow[5],NULL,10);
        strcpy(tg->contact,myrow[6]);
        strcpy(tg->comment,myrow[7]);
        array_init(&(tg->identcounts),sizeof(identcounter));
        tg->currentlyon=0;
        thehash=tgshash(tg->id);
        tg->next=(void *)trustedgroups[thehash];
        trustedgroups[thehash]=tg;
        hostcounter++;
      }
    }
  }
  mysql_free_result(myres);
  hostcounter=hostcounter<<16;
  res=mysql_query(&sqlcon,"SELECT * FROM trustedhosts");
  if (res!=0) {
    sprintf(logline,"!!! Failed to query trustedhosts from database !!!"); putlog();
    return hostcounter;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=2) {
    sprintf(logline,"!!! trustedhosts in database have invalid format !!!"); putlog();
    return hostcounter;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    if (strlen(myrow[0])>HOSTLEN) {
      sprintf(logline,"!!! Malformed trustedhosts-entry in database! !!!"); putlog();
    } else {
      trustedgroup *tg; unsigned int thehash; trustedhost *th;
      tg=findtrustgroupbyID(strtol(myrow[1],NULL,10));
      if (tg==NULL) {
        sprintf(logline,"!!! trustedhosts: Referential integrity violated! (no group #%ld)",strtol(myrow[1],NULL,10));
        putlog();
      } else {
        th=malloc(sizeof(trustedhost));
        strcpy(th->hostname,myrow[0]);
        th->id=strtol(myrow[1],NULL,10);
        thehash=tlhash(myrow[0]);
        th->next=(void *)trustedhosts[thehash];
        trustedhosts[thehash]=th;
        hostcounter++;
      }
    }
  }
  mysql_free_result(myres);
  trustgroupexpire();
  return hostcounter;
}

void savetrusts() {
  char sqlquery[5000]; long i; trustedgroup *tg; trustedhost *th;
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  /* No need to write expired data, so expire old shit first */
  trustgroupexpire();
  /* Clear the tables */
  mysql_query(&sqlcon,"DELETE FROM trustgroups");
  mysql_query(&sqlcon,"DELETE FROM trustedhosts");
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      mysql_escape_string(tmps2,tg->name,strlen(tg->name));
      mysql_escape_string(tmps3,tg->contact,strlen(tg->contact));
      mysql_escape_string(tmps4,tg->comment,strlen(tg->comment));
      sprintf(sqlquery,"INSERT INTO trustgroups VALUES(%lu, '%s', %d, %lu, %d, %d, '%s', '%s')",
        tg->id,tmps2,tg->trustedfor,tg->expires,tg->maxperident,tg->enforceident,tmps3,tmps4);
      mysql_query(&sqlcon,sqlquery);
      tg=(void *)tg->next;
    }
  }
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      mysql_escape_string(tmps2,th->hostname,strlen(th->hostname));
      sprintf(sqlquery,"INSERT INTO trustedhosts VALUES('%s', %lu)",tmps2,th->id);
      mysql_query(&sqlcon,sqlquery);
      th=(void *)th->next;
    }
  }
}

void trustgroupexpire() {
  long i; trustedgroup *tg; unsigned long j;
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      if (tg->expires<getnettime()) {
        j=tg->id;
        tg=(void *)tg->next;
        destroytrustgroup(j);
      } else {
        tg=(void *)tg->next;
      }
    }
  }
}

int trustlistsorthelper(const void *a, const void *b) {
  unsigned long *c, *d;
  c=(unsigned long *)a;
  d=(unsigned long *)b;
  return (*c - *d);
}

void displaytrustgroupwithid(long unum, unsigned long showid) {
  long i; trustedhost *th; trustedgroup *tg; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  tg=findtrustgroupbyID(showid);
  if (tg==NULL) { return; }
  sprintf(tmps2,"Trustgroup '%s' (ID #%lu) - currently online: %d",tg->name,tg->id,tg->currentlyon);
  msgtouser(unum,tmps2);
  longtoduration(tmps3,tg->expires - getnettime());
  sprintf(tmps2,"Settings: Enforcing ident: %s - trusted for %d - %d per user - expires in %s",
    (tg->enforceident == 1) ? "yes" : "no ", tg->trustedfor, tg->maxperident, tmps3);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"Contact: %s - Comment: %s",tg->contact,tg->comment);
  msgtouser(unum,tmps2);
  tmps2[0]=0;
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==showid) {
        if (tmps2[0]==0) {
          strcpy(tmps2,th->hostname);
        } else {
          sprintf(tmps3,"|-%s",tmps2);
          msgtouser(unum,tmps3);
          strcpy(tmps2,th->hostname);
        }
      }
      th=(void *)th->next;
    }
  }
  if (tmps2[0]==0) {
    msgtouser(unum,"`- no hosts in that group");
  } else {
    sprintf(tmps3,"`-%s",tmps2); msgtouser(unum,tmps3);
  }
}

void dotrustlist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  array listofids; long i; trustedgroup *tg; trustedhost *th; long j;
  unsigned long lastshown, showid;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res>2) {
    msgtouser(unum,"Syntax: trustlist [pattern]");
    return;
  }
  if (res==1) { strcpy(tmps3,"*"); }
  toLowerCase(tmps3);
  trustgroupexpire();
  array_init(&listofids,sizeof(unsigned long));
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      strcpy(tmps4,tg->name); toLowerCase(tmps4);
      if (match2strings(tmps3,tmps4)) {
        j=array_getfreeslot(&listofids);
        ((unsigned long *)listofids.content)[j]=tg->id;
      }
      tg=(void *)tg->next;
    }
  }
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      strcpy(tmps4,th->hostname); toLowerCase(tmps4);
      if (match2strings(tmps3,tmps4)) {
        j=array_getfreeslot(&listofids);
        ((unsigned long *)listofids.content)[j]=th->id;
      }
      th=(void *)th->next;
    }
  }
  if (listofids.cursi==0) {
    msgtouser(unum,"No matches.");
    array_free(&listofids);
    return;
  }
  qsort(listofids.content, listofids.cursi, sizeof(unsigned long), trustlistsorthelper);
  lastshown=0;
  for (i=0;i<listofids.cursi;i++) {
    showid=((unsigned long *)(listofids.content))[i];
    if (showid!=lastshown) {
      lastshown=showid;
      displaytrustgroupwithid(unum,showid);
    }
  }
  array_free(&listofids);
}

void dotrustgline(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  char tmps6[TMPSSIZE]; long r2;
  trustedgroup *tg; trustedhost *th; int i; long l; array listofths;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail, "%s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6);
  if (res<4) {
    msgtouser(unum,"Syntax: trustgline <groupname OR #groupID> user duration [reason]");
    msgtouser(unum,"user may contain wildcards");
    return;
  }
  if (res==4) {
    numtonick(unum,tmps2);
    sprintf(tmps6,"requested by %s",tmps2);
  }
  tmps4[USERLEN+2]='\0';
  toLowerCase(tmps3); toLowerCase(tmps4);
  r2=durationtolong(tmps5);
  if (r2<=0) {
    sprintf(tmps2,"The duration you gave is invalid.");
    msgtouser(unum,tmps2);
    return;
  }
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  /* First, create a list of all hosts in that trustgroup... */
  array_init(&listofths,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==tg->id) { /* mmkay, that one belongs to our group */
        long j;
        j=array_getfreeslot(&listofths);
        strcpy(((char *)listofths.content)+(j*(HOSTLEN+1)),th->hostname);
      }
      th=(void *)th->next;
    }
  }
  if (listofths.cursi==0) {
    msgtouser(unum,"There are no hosts in that trustgroup.");
    array_free(&listofths);
    return;
  }
  /* Now that we have the list: count number of users hit */
  l=0;
  for (i=0;i<listofths.cursi;i++) {
    char *bla;
    bla=(char *)listofths.content;
    sprintf(tmps2,"%s@%s",tmps4,bla+(i*(HOSTLEN+1)));
    l+=countusershit(tmps2);
  }
  if (l>500) {
    sprintf(tmps2,"That gline would hit more than 500 (%ld) users. You probably mistyped something.",l);
    msgtouser(unum,tmps2);
    array_free(&listofths); return;
  }
  /* OK, safety checks done - now gline the fuckers */
  numtonick(unum,tmps5);
  for (i=0;i<listofths.cursi;i++) {
    char *bla;
    bla=(char *)listofths.content;
    sprintf(tmps2,"%s@%s",tmps4,bla+(i*(HOSTLEN+1)));
    addgline(tmps2,tmps6,tmps5,r2,1);
  }
  sprintf(tmps2,"Added GLINE for %s @ trustgroup %s (#%lu), caused %lu glines, hit %ld users",tmps4,tg->name,tg->id,
    listofths.cursi,l);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"GLINE %s @ trustgroup %s (#%lu) by %s, caused %lu glines, hit %ld users",tmps4,
    tg->name,tg->id,tmps5,listofths.cursi,l);
  sendtonoticemask(NM_GLINE,tmps2);
  array_free(&listofths);
}

void dotrustspew(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  trustedgroup *tg; trustedhost *th; int i; userdata *ud; long k, l;
  array listofths; int iamtired;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail, "%s %s %s",tmps2,tmps3,tmps4);
  if (res<2) {
    msgtouser(unum,"Syntax: trustspew <groupname OR #groupID> [user]");
    return;
  }
  if (res==2) { strcpy(tmps4,"*"); }
  tmps4[USERLEN+2]='\0';
  toLowerCase(tmps3); toLowerCase(tmps4);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  /* First, create a list of all hosts in that trustgroup... */
  array_init(&listofths,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==tg->id) { /* mmkay, that one belongs to our group */
        long j;
        j=array_getfreeslot(&listofths);
        strcpy(((char *)listofths.content)+(j*(HOSTLEN+1)),th->hostname);
      }
      th=(void *)th->next;
    }
  }
  if (listofths.cursi==0) {
    msgtouser(unum,"There are no hosts in that trustgroup.");
    array_free(&listofths);
    return;
  }
  /* We have the list - scan through the whole userlist, and match against our
     hostlist, and if that matches, match again against the username */
  sprintf(tmps2,"Users in Trustgroup %s (#%lu) matching %s",tg->name,tg->id,tmps4);
  msgtouser(unum,tmps2);
  l=0;
  for (i=0;i<SIZEOFUL;i++) {
    ud=uls[i];
    while (ud!=NULL) {
      strcpy(tmps2,ud->host);
      toLowerCase(tmps2);
      for (k=0;k<listofths.cursi;k++) {
        if (strcmp(tmps2,((char *)listofths.content)+(k*(HOSTLEN+1)))==0) {
          /* Host matches, check username */
          strcpy(tmps5,ud->ident);
          toLowerCase(tmps5);
          if (match2strings(tmps4,tmps5)) {
            /* yay ident matches too - list that loser */
            channel **d; int chanmodes;
            l++; /* inc counter of number of users found */
            sprintf(tmps2,"%s!%s@%s (%s)",ud->nick,ud->ident,ud->host,ud->realname);
            msgtouser(unum,tmps2);
            if (ud->chans.cursi>0) {
              if (ud->chans.cursi==1) {
                strcpy(tmps2,"On channel:");
              } else {
                strcpy(tmps2,"On channels:");
              }
              for (iamtired=0;iamtired<(ud->chans.cursi);iamtired++) {
                d=(channel **)(ud->chans.content);
                if ((strlen(d[iamtired]->name)+strlen(tmps2))>400) {
                  strcat(tmps2," [...]"); break;
                } else {
                  chanmodes=getchanmode2(d[iamtired],ud->numeric);
                  strcat(tmps2," (");
                  if (chanmodes<0) {
                    strcat(tmps2,"??");
                  } else {
                    if (isflagset(chanmodes,um_o)) { strcat(tmps2,"@"); } else { strcat(tmps2,"_"); }
                    if (isflagset(chanmodes,um_v)) { strcat(tmps2,"+"); } else { strcat(tmps2,"_"); }
                  }
                  strcat(tmps2,")");
                  strcat(tmps2,d[iamtired]->name);
                }
              }
            } else {
              strcpy(tmps2,"Not on any (nonlocal-)channels");
            }
            msgtouser(unum,tmps2);
          }
        }
      }
      ud=(void *)ud->next;
    }
  }
  sprintf(tmps2,"--- End of list - %ld matches ---",l);
  msgtouser(unum,tmps2);
  array_free(&listofths);
}
