/*
Operservice 2 - usercommands.c
Supplies all the standard commands that users can /msg the service
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "mysql_version.h"
#include "mysql.h"
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <malloc.h>
#include "globals.h"

void dowhois(long unum, char *tail) {
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; int um;
  int res; userdata **a; long i; unsigned int hash; channel **d; int iamtired;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if ((res!=2) || (strcmp(tmps3,"")==0)) {
    msgtouser(unum,"Syntax: whois nickname");
    return;
  }
  toLowerCase(tmps3);
  hash=nlhash(tmps3);
  a=(userdata **)nls[hash].content;
  for (i=0;i<nls[hash].cursi;i++) {
    if (strcmp(tmps3,a[i]->nick)==0) {
      sprintf(tmps2,"WHOIS info for %s:",tmps3);
      msgtouser(unum,tmps2);
      sprintf(tmps2,"Usermodes:        %s",a[i]->umode);
      msgtouser(unum,tmps2);
      sprintf(tmps2,"Ident@Host:       %s@%s",a[i]->ident,a[i]->host);
      msgtouser(unum,tmps2);
      longtotoken(a[i]->numeric,tmps3,5);
      sprintf(tmps2,"Numeric:          %s",tmps3);
      msgtouser(unum,tmps2);
      sprintf(tmps2,"Realname:         %s",a[i]->realname);
      msgtouser(unum,tmps2);
      longtoduration(tmps3,getnettime()-a[i]->connectat);
      sprintf(tmps2,"Connected for:    %s",tmps3);
      msgtouser(unum,tmps2);
      sprintf(tmps2,"Authlevel:        %d",getauthlevel(a[i]->numeric));
      msgtouser(unum,tmps2);
      if (getauthlevel(a[i]->numeric)>0) {
        getauthedas(tmps3,a[i]->numeric);
        sprintf(tmps2,"Authed as:        %s",tmps3);
        msgtouser(unum,tmps2);
      } else {
        msgtouser(unum,"Not Authed");
      }
      if (a[i]->chans.cursi>0) {
        if (a[i]->chans.cursi==1) {
          strcpy(tmps2,"On channel:");
        } else {
          strcpy(tmps2,"On channels:");
        }
        for (iamtired=0;iamtired<(a[i]->chans.cursi);iamtired++) {
          d=(channel **)(a[i]->chans.content);
          if ((strlen(d[iamtired]->name)+strlen(tmps2))>400) {
            strcat(tmps2," [...]"); break;
          } else {
            um=getchanmode2(d[iamtired],a[i]->numeric);
            strcat(tmps2," (");
            if (um<0) {
              strcat(tmps2,"??");
            } else {
              if (isflagset(um,um_o)) { strcat(tmps2,"@"); } else { strcat(tmps2,"_"); }
              if (isflagset(um,um_v)) { strcat(tmps2,"+"); } else { strcat(tmps2,"_"); }
            }
            strcat(tmps2,")");
            strcat(tmps2,d[iamtired]->name);
          }
        }
        msgtouser(unum,tmps2);
      } else {
        msgtouser(unum,"Not on any (nonlocal-)channels");
      }
      msgtouser(unum,"--- End of WHOIS ---");
      return;
    }
  }
  sprintf(tmps2,"User %s does not exist.",tmps3);
  msgtouser(unum,tmps2);
}

void dostatus(long unum, char *tail) {
  int usersonnet; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; int authl;
  int n; long snltot=0; long snluse=0;
  usersonnet=countusersonthenet();
  sprintf(tmps2,"There are currently %d users on the network (%lu fakeusers).",usersonnet,myfakes.cursi);
  msgtouser(unum,tmps2);
  for (n=0;n<256;n++) {
    if (mainsnl[n].cursize!=0) { snluse++; }
    snltot+=mainsnl[n].cursize;
  }
  sprintf(tmps2,"Subnetlistsize is %ld; %ld subnets trusted.",snltot,tsncu);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"Current network time: %lu",(unsigned long)getnettime());
  msgtouser(unum,tmps2);
  longtoduration(tmps3,(time(NULL)-starttime));
  sprintf(tmps2,"I am running for %s",tmps3);
  msgtouser(unum,tmps2);
  authl=getauthlevel(unum);
  if (authl>=999) {
    int i, j;
#ifdef HAVE_MALLINFO
    struct mallinfo mi;
#endif
    j=0;
    for (i=0;i<SIZEOFCL;i++) {
      if (cls[i]!=NULL) { j++; }
    }
    sprintf(tmps2,"Channel-Hashtable-Usage: %d / %d (%5.1f%%)",j,SIZEOFCL,j*100.0/SIZEOFCL);
    msgtouser(unum,tmps2);
    j=0;
    for (i=0;i<SIZEOFUL;i++) {
      if (uls[i]!=NULL) { j++; }
    }
    sprintf(tmps2,"Userlist-Hashtable-Usage: %d / %d (%5.1f%%)",j,SIZEOFUL,j*100.0/SIZEOFUL);
    msgtouser(unum,tmps2);
    j=0;
    for (i=0;i<SIZEOFNL;i++) {
      if (nls[i].cursi>0) { j++; }
    }
    sprintf(tmps2,"Nicklist-Hashtable-Usage: %d / %d (%5.1f%%)",j,SIZEOFNL,j*100.0/SIZEOFNL);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Subnetlists: %ld / 256 in use (%5.1f%%)",snluse,snluse*100.0/256.0);
    msgtouser(unum,tmps2);
#ifdef HAVE_MALLINFO
    mi=mallinfo();
    sprintf(tmps2,"malloc(): %d used, %d unused, %5.1f%% waste",mi.uordblks,mi.fordblks,((100.0*mi.fordblks)/(mi.fordblks+mi.uordblks)));
    msgtouser(unum,tmps2);
#else
    msgtouser(unum,"malloc(): compiled without mallinfo(), no information available.");
#endif
  }
}

void doopchan(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  long tmpl;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s",tmps2,tmps3,tmps4,tmps5);
  if (res!=3) {
    sprintf(tmps2,"Syntax: opchan channel nickname");
    msgtouser(unum,tmps2);
    return;
  }
  toLowerCase(tmps4); toLowerCase(tmps3);
  tmpl=nicktonum(tmps4);
  if (tmpl<0) {
    sprintf(tmps2,"User %s is not on the network.",tmps4);
    msgtouser(unum,tmps2);
    return;
  }
  longtotoken(tmpl,tmps2,5);
  fprintf(sockout,"%s M %s +o %s\r\n",servernumeric,tmps3,tmps2);
  sprintf(tmps2,"Put fake mode +o %s on %s",tmps4,tmps3);
  changechanmode(tmps3,tmpl,1,um_o);
  msgtouser(unum,tmps2);
  fflush(sockout);
}

void dokickcmd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], reas[TMPSSIZE];
  long tmpl;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %[^\n]",tmps2,tmps3,tmps4,reas);
  if (res<3) {
    msgtouser(unum,"Syntax: kick channel nickname [reason]");
    msgtouser(unum,"If you don't give a reason the default will be \"requested by (yournick)\"");
    return;
  }
  if (res==3) {
    numtonick(unum,tmps2);
    sprintf(reas,"requested by %s",tmps2);
  }
  toLowerCase(tmps4); toLowerCase(tmps3);
  tmpl=nicktonum(tmps4);
  if (tmpl<0) {
    sprintf(tmps2,"User %s is not on the network.",tmps4);
    msgtouser(unum,tmps2);
    return;
  }
  longtotoken(tmpl,tmps2,5);
  fprintf(sockout,"%s K %s %s :%s\r\n",servernumeric,tmps3,tmps2,reas);
  sprintf(tmps2,"Put fake kick for %s on %s (Reason: %s)",tmps4,tmps3,reas);
  deluserfromchan(tmps3,tmpl);
  msgtouser(unum,tmps2);
  fflush(sockout);
}

void dojupe(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE]; long tmpl;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %ld %s",tmps2,tmps3,&tmpl,tmps4);
  if (res!=3) {
    sprintf(tmps2,"Syntax: jupe servername numeric");
    msgtouser(unum,tmps2);
    return;
  }
  if (numericinuse(tmpl)) {
    if (getsrvnum(tmps3)!=tmpl) {
      sprintf(tmps2,"Sorry, the numeric you supplied is already in use on the network.");
      msgtouser(unum,tmps2);
      return;
    }
  }
  if (getsrvnum(tmps3)>=0) {
    fprintf(sockout,"%s SQ %s 0 :Getting juped\r\n",servernumeric,tmps3);
    delsrv(tmps3);
  }
  longtotoken(tmpl,tmps2,2);
  fprintf(sockout,"%s S %s 2 0 %ld J10 %sAAB :Jupe - /SQUIT to remove\r\n",servernumeric,tmps3,getnettime()+84000,tmps2);
  fprintf(sockout,"%s EB\r\n",tmps2);
  fprintf(sockout,"%s EA\r\n",tmps2);
  newserver(tmps3,1,tmpl,0,getnettime(),tokentolong(servernumeric),1);
  sprintf(tmps4,"%s has been juped.",tmps3);
  msgtouser(unum,tmps4);
  fflush(sockout);
}

void dooperhelp(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res>1) {
    showhelp(unum,tmps3);
    return;
  }
  msgtouser(unum,"Additional commands for opers:");
  msgtouser(unum,"-----------------------------------------------------------------------");
  msgtouser(unum,"hello [desired username]  Creates a new useraccount (with zero rights)");
  msgtouser(unum,"password newpass newpass  Changes your password");
  msgtouser(unum,"noticeme on|off           Sets if the service sends you its replys as /msg or as /notice");
  msgtouser(unum,"status                    Prints some status information");
  msgtouser(unum,"jupe servername numeric   Jupes a server (prevents it from connecting to the net)");
  msgtouser(unum,"opchan channel nickname   Ops nickname on channel");
  msgtouser(unum,"kick chan nick [reason]   Kicks nickname from channel");
  msgtouser(unum,"whois nickname            Shows some info about this user");
  msgtouser(unum,"channel channelname       Shows info about that channel");
  msgtouser(unum,"resync channelname        Resyncs a desynched channel");
  msgtouser(unum,"semiresync channelname    Resyncs a desynched channel");
  msgtouser(unum,"deopall channelname       Deops everyone on a channel (including Q!)");
  msgtouser(unum,"glist [pattern]           Lists glines");
  msgtouser(unum,"gline hostm dur [reas]    Sets a gline");
  msgtouser(unum,"ungline hostmask          Removes a gline");
  msgtouser(unum,"syncglines                Synchs glines on all servers");
  msgtouser(unum,"countusers [pattern]      Counts how many Users on the net match pattern");
  msgtouser(unum,"mfc [timespan]            Mirkforce-Check");
  msgtouser(unum,"serverlist                Lists servers");
  msgtouser(unum,"password [user] newp newp Changes password for you or another user");
  msgtouser(unum,"changelev user newlevel   Changes the authlevel of a user");
  msgtouser(unum,"deluser username          Deletes a user");
  msgtouser(unum,"broadcast message         Broadcasts a message to all users");
  msgtouser(unum,"sbroadcast srv msg        Broadcasts a message to all users on a specific server");
  msgtouser(unum,"listauthed                Lists who is authed right now");
  msgtouser(unum,"subnetlist X.Y.Z          Lists all users on a subnet");
  msgtouser(unum,"sntrustlist [pattern]     Lists trusted subnets");
  msgtouser(unum,"sntrustadd X.Y.Z numusers Adds a subnet as trusted");
  msgtouser(unum,"          duration reason");
  msgtouser(unum,"sntrustdel X.Y.Z          Removes a trusted subnet");
  msgtouser(unum,"spew host|searchmask      Lists all clients on a certain host or matching a searchmask");
  msgtouser(unum,"fakeuser nick ident       Creates a fake user (useful to block nicks)");
  msgtouser(unum,"          host realname");
  msgtouser(unum,"fakelist                  Lists all Fakeusers");
  msgtouser(unum,"fakekill nick             Removes a fakeuser");
  msgtouser(unum,"rnc                       Lists the 20 most common realnames on the net");
  msgtouser(unum,"rngline pattern duration  Adds a GLINE on a Realname");
  msgtouser(unum,"          bantype reason");
  msgtouser(unum,"rnungline pattern         Removes a GLINE on a Realname");
  msgtouser(unum,"rnglist                   Lists all Realname-GLINEs");
  msgtouser(unum,"splitlist                 Gives a list of splits");
  msgtouser(unum,"splitdel pattern          Removes one or more entries from the splitlist");
  msgtouser(unum,"loadmod module            Loads a module");
  msgtouser(unum,"unloadmod module          Unloads a module");
  msgtouser(unum,"reloadmod module          Same as unloadmod followed by loadmod");
  msgtouser(unum,"lsmod                     Lists loaded modules");
  msgtouser(unum,"noticemask [newmask]      Shows/Sets your noticemask");
  msgtouser(unum,"die [message]             Kills the service. Does NOT automatically save before that!");
  msgtouser(unum,"trustgroupadd ...         Adds a new or mofifies an existing trustgroup");
  msgtouser(unum,"trustgroupdel name|#gid   Deletes a trustgroup (be careful with that!)");
  msgtouser(unum,"trustadd name|#gid host   Adds a host to a trustgroup");
  msgtouser(unum,"trustdel host             Removes a host from a trustgroup");
  msgtouser(unum,"trustlist [pattern]       Lists matching or all trustgroups/hosts");
  msgtouser(unum,"trustgline ...            GLINEs a user on all hosts of a trustgroup");
  msgtouser(unum,"trustspew group [user]    Lists users on a trustgroup");
  dyncmdhelp(unum,1);
}

void douserhelp(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res>1) {
    sprintf(tmps4,"user%s",tmps3);
    showhelp(unum,tmps4);
    return;
  }
  msgtouser(unum,"Available commands: (not all commands are necessarily available to you)");
  msgtouser(unum,"-----------------------------------------------------------------------");
  msgtouser(unum,"help [command]            Help for a specific command");
  msgtouser(unum,"auth username password    auths you");
  dyncmdhelp(unum,0);
}

void dochancmd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  channel *a; chanuser *b; int ops; int voices;
  unsigned long j; int k, l;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: channel channelname");
    return;
  }
  toLowerCase(tmps3);
  a=getchanptr(tmps3);
  if (a==NULL) {
    msgtouser(unum,"That channel does not exist.");
    return;
  }
  sprintf(tmps2,"Information about %s",tmps3);
  msgtouser(unum,tmps2);
  strcpy(tmps2,""); ops=0;
  for (res=0;res<numcf;res++) {
    if (isflagset(a->flags,chanflags[res].v)) {
      tmps2[ops]=chanflags[res].c;
      ops++;
    }
  }
  tmps2[ops]='\0';
  sprintf(tmps4,"Chanmodes: %s",tmps2);
  msgtouser(unum,tmps4);
  if (isflagset(a->flags,cm_l)) {
    sprintf(tmps4,"Channellimit: %d",a->chanlim);
    msgtouser(unum,tmps4);
  }
  if (isflagset(a->flags,cm_k)) {
    sprintf(tmps4,"Channelkey: %s",a->chankey);
    msgtouser(unum,tmps4);
  }
  if (strcmp(a->topic,"")==0) {
    msgtouser(unum,"No topic set");
  } else {
    sprintf(tmps2,"Topic: %s",a->topic);
    msgtouser(unum,tmps2);
  }
  msgtouser(unum,"Users on the channel:");
  b=(chanuser *)a->users.content; strcpy(tmps2,""); res=0; ops=0; voices=0;
  for (j=0;j<a->users.cursi;j++) {
    if (strlen(tmps2)>80) {
      msgtouser(unum,tmps2);
      strcpy(tmps2,"");
    }
    numtonick(b[j].numeric,tmps4);
    l=strlen(tmps4);
    if (l>0) {
      for (k=l;k<NICKLEN;k++) {
        tmps4[k]=' ';
      }
      tmps4[NICKLEN]='\0';
    }
    if (tmps2[0]!='\0') { strcat(tmps2," "); }
    strcat(tmps2,"(");
    if (isflagset(b[j].flags,um_o)) {
      ops++;
      strcat(tmps2,"@");
    } else {
      strcat(tmps2,"_");
    }
    if (isflagset(b[j].flags,um_v)) {
      voices++;
      strcat(tmps2,"+");
    } else {
      strcat(tmps2,"_");
    }
    strcat(tmps2,")");
    if (tmps4[0]=='\0') {
      longtotoken(b[j].numeric,tmps4,5);
      strcat(tmps2,"!! DEBUG: Empty nick. Numeric: ");
      strcat(tmps2,tmps4);
      strcat(tmps2," !!");
    } else {
      strcat(tmps2,tmps4);
    }
    res++;
  }
  msgtouser(unum,tmps2);
  sprintf(tmps2,"%d Users; %d Op, %d Voice",res,ops,voices);
  msgtouser(unum,tmps2);
}

void doresync(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  channel *a; chanuser *b; unsigned long j;
  if (!checkauthlevel(unum,500)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: resync channelname");
    return;
  }
  toLowerCase(tmps3);
  a=getchanptr(tmps3);
  if (a==NULL) {
    msgtouser(unum,"That channel does not exist.");
    return;
  }
  b=(chanuser *)a->users.content; strcpy(tmps2,""); res=0; strcpy(tmps4,"");
  for (j=0;j<a->users.cursi;j++) {
    if (res>=6) {
      fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4);
      res=0; strcpy(tmps2,""); strcpy(tmps4,"");
    }
    if (isflagset(b[j].flags,um_o)) {
      strcat(tmps2,"+o");
    } else {
      strcat(tmps2,"-o");
    }
    longtotoken(b[j].numeric,tmps5,5);
    if (res>0) { strcat(tmps4," "); }
    strcat(tmps4,tmps5);
    res++;
  }
  fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4);
  msgtouser(unum,"Channel should be resynced.");
  fflush(sockout);
}

void dosemiresync(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  channel *a; chanuser *b; unsigned long j;
  if (!checkauthlevel(unum,500)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: semiresync channelname");
    return;
  }
  toLowerCase(tmps3);
  a=getchanptr(tmps3);
  if (a==NULL) {
    msgtouser(unum,"That channel does not exist.");
    return;
  }
  b=(chanuser *)a->users.content; strcpy(tmps2,""); res=0; strcpy(tmps4,"");
  for (j=0;j<a->users.cursi;j++) {
    if (res>=6) {
      fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4);
      res=0; strcpy(tmps2,""); strcpy(tmps4,"");
    }
    if (isflagset(b[j].flags,um_o)) {
      strcat(tmps2,"+o");
      longtotoken(b[j].numeric,tmps5,5);
      if (res>0) { strcat(tmps4," "); }
      strcat(tmps4,tmps5);
      res++;
    }
  }
  if (res>0) { fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4); }
  msgtouser(unum,"Channel should be semiresynced.");
  fflush(sockout);
}

void dodeopall(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  channel *a; chanuser *b; unsigned long j;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: deopall channelname");
    return;
  }
  toLowerCase(tmps3);
  a=getchanptr(tmps3);
  if (a==NULL) {
    msgtouser(unum,"That channel does not exist.");
    return;
  }
  b=(chanuser *)a->users.content; strcpy(tmps2,""); res=0; strcpy(tmps4,"");
  for (j=0;j<a->users.cursi;j++) {
    if (res>=6) {
      fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4);
      res=0; strcpy(tmps2,""); strcpy(tmps4,"");
    }
    strcat(tmps2,"-o");
    longtotoken(b[j].numeric,tmps5,5);
    if (res>0) { strcat(tmps4," "); }
    strcat(tmps4,tmps5);
    res++;
    changechanmod2(a,b[j].numeric,2,um_o);
  }
  fprintf(sockout,"%s M %s %s %s\r\n",servernumeric,tmps3,tmps2,tmps4);
  msgtouser(unum,"deopall should be done.");
  fflush(sockout);
}

void doglist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; int i;
  MYSQL_RES *myres; MYSQL_ROW myrow;
  purgeglines();
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res<1) {
    msgtouser(unum,"Syntax: glist [pattern]");
    return;
  }
  if (res<2) { strcpy(tmps3,"%"); }
  for (i=0;i<strlen(tmps3);i++) {
    if (tmps3[i]=='*') { tmps3[i]='%'; }
  }
  mysql_escape_string(tmps4,tmps3,strlen(tmps3));
  sprintf(tmps5,"SELECT * FROM glines WHERE hostmask LIKE '%s'",tmps4);
  res=mysql_query(&sqlcon,tmps5);
  if (res!=0) {
    sprintf(logline,"!!! Failed to query gline-list from database !!!"); putlog();
    msgtouser(unum,logline);
    return;
  }
  sprintf(tmps2,"GLINEs matching %s:",tmps3);
  msgtouser(unum,tmps2); i=0;
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=4) {
    mysql_free_result(myres);
    sprintf(logline,"!!! GList in database has invalid format !!!"); putlog();
    msgtouser(unum,logline);
    return;
  }
  sprintf(tmps2,"%-50s %-17s by %s: %s","Hostmask","expires in","whom","reason");
  msgtouser(unum,tmps2);
  while ((myrow=mysql_fetch_row(myres))) {
    longtoduration(tmps5,atol(myrow[2])-getnettime());
    sprintf(tmps2,"%-50s %-17s by %s: %s",myrow[0],tmps5,myrow[1],myrow[3]);
    msgtouser(unum,tmps2);
    i++;
  }
  sprintf(tmps2,"--- End of list - %d matches ---",i);
  msgtouser(unum,tmps2);
  mysql_free_result(myres);
}

void dogline(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE]; int r2;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5);
  if (res<3) {
    msgtouser(unum,"Syntax: gline hostmask duration [reason]");
    return;
  }
  if (res==3) {
    numtonick(unum,tmps2);
    sprintf(tmps5,"requested by %s",tmps2);
  }
  r2=durationtolong(tmps4);
  if (r2<=0) {
    sprintf(tmps2,"The duration you gave is invalid.");
    msgtouser(unum,tmps2);
    return;
  }
  res=countusershit(tmps3);
  if (res>500) {
    sprintf(tmps2,"That gline would hit more than 500 (%d) Users/Channels. You probably misstyped something.",res);
    msgtouser(unum,tmps2);
    return;
  }
  numtonick(unum,tmps6);
  sprintf(tmps2,"GLINE %s, expires in %s, set by %s: %s (hit %d %s%s)",tmps3,tmps4,tmps6,tmps5,res,(tmps3[0]=='#') ? "channel" : "user",(res==1) ? "" : "s");
  sendtonoticemask(NM_GLINE,tmps2);
  purgeglines();
  addgline(tmps3,tmps5,tmps6,r2,1);
  sprintf(tmps2,"Added GLINE %s, expires in %d seconds, hit %d %s%s.",tmps3,r2,res,(tmps3[0]=='#') ? "channel" : "user",(res==1) ? "" : "s");
  msgtouser(unum,tmps2);
}

void doungline(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: ungline hostmask");
    return;
  }
  numtonick(unum,tmps4);
  sprintf(tmps2,"GLINE %s removed by %s",tmps3,tmps4);
  sendtonoticemask(NM_GLINE,tmps2);
  remgline(tmps3,1);
  msgtouser(unum,"Done.");
}

void docountusers(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) { strcpy(tmps3,"*"); }
  res=countusershit(tmps3);
  sprintf(tmps2,"%d users match %s",res,tmps3);
  msgtouser(unum,tmps2);
}

void domfc(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; long since; int i;
  subnetlist mfcl; userdata *a; subnetdata *p;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    since=0;
  } else {
    since=getnettime()-atol(tmps3);
  }
  snlcreate(&mfcl,1000);
  for (i=0;i<SIZEOFUL;i++) {
    a=uls[i];
    while (a!=NULL) {
      if (a->connectat>=since) { snladd(&mfcl,a->realip); }
      a=(void *)a->next;
    }
  }
  snlsort(&mfcl);
  p=mfcl.s;
  if (since==0) {
    msgtouser(unum,"Top 20: subnets with the most connections");
  } else {
    sprintf(tmps2,"Top 20: subnets with the most connections since %ld (networktime)",since);
    msgtouser(unum,tmps2);
  }
  for (res=0;res<20;res++) {
    if (res<mfcl.cursize) {
      sprintf(tmps2,"%3d.%3d.%3d.*  %4d",p->a,p->b,p->c,p->n);
      msgtouser(unum,tmps2);
    }
    p++;
  }
  snlfree(&mfcl);
}

void dosettimecmd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  if (!checkauthlevel(unum,500)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if ((res==2) && (strcmp(tmps3,"force")==0)) {
    msgtouser(unum,"Sending settime (forced)");
    timestdif=0;  /* _WE_ are authoritative with respect to time. */
    fprintf(sockout,"%s SE %ld\r\n",servernumeric,(long int)getnettime());
    fflush(sockout);
  } else {
    if (waitingforping==0) {
      msgtouser(unum,"Sending pings to all servers. If the networklag is below a certain value, i will then do SETTIME.");
      msgtouser(unum,"If you want to force a settime even if the net is lagged, use 'settime force'.");
      pingallservers();
      waitingforping=getnettime(); lastsettime=getnettime()-3600;
    } else {
      msgtouser(unum,"There is already a settime-request pending.");
    }
  }
}

void doserverlist(long unum, char *tail) {
  serverdata *s; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE]; int sc;
  long usersonthissrv; long totalusers=0;
  sprintf(tmps2,"%3s %-32s %-18s %-6s %-18s %4s %4s %s","Num","Hostname","Uptime","MaxCl","Connected for","Ping","Clie","Ports");
  msgtouser(unum,tmps2);
  s=sls; sc=0;
  while (s!=NULL) {
    sc++;
    longtoduration(tmps4,getnettime()-s->connectat);
    longtoduration(tmps3,getnettime()-s->createdat);
    if ((getnettime()-s->createdat)>durationtolong("3y")) {
      strcpy(tmps3,"N/A");
    }
    usersonthissrv=countusersonsrv(s->numeric);
    totalusers+=usersonthissrv;
    if ((s->pingreply-s->pingsent)<0) {
      sprintf(tmps2,"%3ld %-32s %18s %6ld %18s  N/A %4ld %s",s->numeric,s->nick,tmps3,s->maxclients,tmps4,usersonthissrv,s->ports);
    } else {
      sprintf(tmps2,"%3ld %-32s %18s %6ld %18s %4ld %4ld %s",s->numeric,s->nick,tmps3,s->maxclients,tmps4,s->pingreply-s->pingsent,usersonthissrv,s->ports);
    }
    msgtouser(unum,tmps2);
    s=(void *)s->next;
  }
  longtoduration(tmps3,getnettime()-starttime);
  usersonthissrv=countusersonsrv(tokentolong(servernumeric));
  totalusers+=usersonthissrv;
  sprintf(tmps2,"%3ld %-32s %18s %6d %18s %4d %4ld %s",
    tokentolong(servernumeric),myname,tmps3,4095,tmps3,0,usersonthissrv,"none");
  msgtouser(unum,tmps2);
  sprintf(tmps2,"--- End of list. %ld users and %d servers on the net.",totalusers,sc+1);
  msgtouser(unum,tmps2);
}

void dohello(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  authdata ad;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res<2) { numtonick(unum,tmps3); }
  tmps3[32]='\0';
  getuserdata(&ad,tmps3);
  if (ad.authlevel<0) {
    msgtouser(unum,"Error accessing the database. Try again later.");
    return;
  }
  if (ad.authlevel>0) {
    sprintf(tmps2,"Sorry, Username %s is already in use.",tmps3);
    msgtouser(unum,tmps2);
    msgtouser(unum,"You'll have to use another Username. Try 'hello otherusername'");
    return;
  }
  // That user does not exist yet, so lets create it!
  strcpy(ad.username,tmps3);
  createrandompw(ad.password,8);
  sprintf(tmps2,"A random password has been created for you: %s",ad.password);
  msgtouser(unum,tmps2);
  msgtouser(unum,"Please change this password as soon as possible.");
  if (countusersinul()==0) {
    ad.authlevel=1000;  // This is the first user on the service, he gets assigned level 1000
    msgtouser(unum,"You are the first user on the service. You have been assigned an authlevel of 1000.");
    strcpy(tmps2,ad.password);      // We need to encrypt this users PW
    encryptpwd(ad.password,tmps2);  // because he has authlevel >=100 !
  } else {
    ad.authlevel=1;
  }
  ad.lastauthed=getnettime();
  ad.wantsnotice=0;
  if (isircop(unum)) { ad.noticemask=NOTICEMASKDEFAULT; } else { ad.noticemask=0; }
  res=updateuserinul(ad);
  if (res!=1) {
    sprintf(tmps2,"Oooops... Database-Query failed. try again later.");
    msgtouser(unum,tmps2);
  } else {
    sprintf(tmps2,"You have been successfully added with username %s",ad.username);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"You can now auth: /msg %s auth %s yourpassword",mynick,ad.username);
    msgtouser(unum,tmps2);
  }
}

void doauth(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; authdata ad;
  char lohlp[TMPSSIZE];
  res=sscanf(tail,"%s %s %s %s",tmps2,tmps3,tmps4,tmps5);
  if (res!=3) {
    msgtouser(unum,"Syntax: auth username password");
    msgtouser(unum,"Note that password is case sensitive, username isn't.");
    return;
  }
  numtohostmask(unum,lohlp);
  tmps3[32]='\0'; tmps4[32]='\0';
  getuserdata(&ad,tmps3);
  if (ad.authlevel<0) {
    msgtouser(unum,"Sorry, Query to database failed. Please try again later.");
    sprintf(logline,"%s !AUTH %s! - failed (DB)",lohlp,tmps3); putlog();
    return;
  }
  if (ad.authlevel==0) {
    msgtouser(unum,"That user does not exist.");
    sprintf(logline,"%s !AUTH %s! - failed (no such user)",lohlp,tmps3); putlog();
    return;
  }
  if (ad.authlevel>=100) {
    if (checkpwd(ad.password,tmps4)!=1) {
      msgtouser(unum,"Password incorrect.");
      sprintf(logline,"%s !AUTH %s! - failed (wrong PW)",lohlp,tmps3); putlog();
      return;
    }
  } else {
    if (strcmp(ad.password,tmps4)!=0) {
      msgtouser(unum,"Password incorrect.");
      sprintf(logline,"%s !AUTH %s! - failed (wrong PW)",lohlp,tmps3); putlog();
      return;
    }
  }
  addautheduser(unum,ad);
  ad.lastauthed=getnettime();
  updateuserinul(ad);
  msgtouser(unum,"Auth successfull.");
  sprintf(logline,"%s !AUTH! %s - succeeded",lohlp,tmps3); putlog();
  if (ad.authlevel>=100) {
    sprintf(tmps2,"Your authlevel is %d",ad.authlevel);
    msgtouser(unum,tmps2);
  }
}

void dochangepwd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE];
  authdata ad; char lohlp[TMPSSIZE];
  res=sscanf(tail,"%s %s %s %s %s",tmps2,tmps3,tmps4,tmps5,tmps6);
  if ((res!=3) && (res!=4)) {
    msgtouser(unum,"Syntax: password [user] newpass newpass");
    return;
  }
  tmps3[32]='\0'; tmps4[32]='\0'; tmps5[32]='\0'; //Truncate everything to 32 chars
  numtohostmask(unum,lohlp);
  if (res==4) {
    if (!isircop(unum)) {
      msgtouser(unum,"Only ircops can change the passwords of other users");
      sprintf(logline,"%s !PASSWORD %s! - failed (no IRCOP)",lohlp,tmps3); putlog();
      return;
    }
    getuserdata(&ad,tmps3);
    if (ad.authlevel>=getauthlevel(unum)) {
      msgtouser(unum,"You cannot change the password of a user that has a higher or equal authlevel");
      sprintf(logline,"%s !PASSWORD %s! - failed (authlevel too low)",lohlp,tmps3); putlog();
      return;
    }
    if (strcmp(tmps4,tmps5)!=0) {
      msgtouser(unum,"Passwords do not match");
      return;
    }
    strcpy(tmps6,tmps3);
    strcpy(tmps2,tmps4);
  } else {
    if (strcmp(tmps3,tmps4)!=0) {
      msgtouser(unum,"Passwords do not match");
      return;
    }
    getauthedas(tmps6,unum);
    strcpy(tmps2,tmps4);
  }
  getuserdata(&ad,tmps6);
  if (ad.authlevel==0) {
    msgtouser(unum,"That user does not exist.");
    sprintf(logline,"%s !PASSWORD %s! - failed (nosuchuser)",lohlp,tmps6); putlog();
    return;
  }
  if (ad.authlevel<0) {
    msgtouser(unum,"There was an error accessing the databse.");
    sprintf(logline,"%s !PASSWORD %s! - failed (DB-Err)",lohlp,tmps6); putlog();
    return;
  }
  if (ad.authlevel>=100) {
    strcpy(tmps5,tmps2);
    encryptpwd(tmps2,tmps5);
  }
  strcpy(ad.password,tmps2);
  if (updateuserinul(ad)) {
    msgtouser(unum,"Password changed.");
    sprintf(logline,"%s !PASSWORD %s! - succeeded",lohlp,tmps6); putlog();
  } else {
    msgtouser(unum,"Something went wrong when accessing the database. Password could not be changed.");
    sprintf(logline,"%s !PASSWORD %s! - failed (DB-Err #2)",lohlp,tmps6); putlog();
  }
}

void dochangelev(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; int newlev;
  authdata ad;
  res=sscanf(tail,"%s %s %s %s",tmps2,tmps3,tmps4,tmps5);
  if (res!=3) {
    msgtouser(unum,"Syntax: changelev user newlevel");
    return;
  }
  newlev=atoi(tmps4);
  getuserdata(&ad,tmps3);
  if (ad.authlevel<0) {
    msgtouser(unum,"Ooooops... Error accessing database.");
    return;
  }
  if (ad.authlevel==0) {
    msgtouser(unum,"That user does not exist.");
    return;
  }
  if (ad.authlevel>=getauthlevel(unum)) {
    msgtouser(unum,"You cannot change the authlevel of a user that has a higher or equal authlevel.");
    return;
  }
  if (newlev>=getauthlevel(unum)) {
    msgtouser(unum,"You cannot give anyone an authlevel that is higher or equal than your own!");
    return;
  }
  if ((newlev<100) && (ad.authlevel>=100)) {
    msgtouser(unum,"*MEEP* You cannot change the level of a user from 100+ to something <100");
    msgtouser(unum,"You'll have to delete that user and recreate it.");
    msgtouser(unum,"The reason for this is, that for users with level >= 100 only a password-hash is stored, not the password itself, and there is no way to get the password back.");
    return;
  }
  if ((ad.authlevel<100) && (newlev>=100)) {
    strcpy(tmps2,ad.password);
    encryptpwd(ad.password,tmps2);
  }
  ad.authlevel=newlev;
  if (updateuserinul(ad)) {
    msgtouser(unum,"Userlevel changed. Note that this change will not take effect until the user reauths.");
  } else {
    msgtouser(unum,"Couldn't change the entry in the database. Try again later.");
  }
}

void dodelusercmd(long unum, char *tail) {
  int res; authdata ad; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: deluser username");
    return;
  }
  getuserdata(&ad,tmps3);
  if (ad.authlevel==0) {
    msgtouser(unum,"That user does not exist");
    return;
  }
  if (ad.authlevel<0) {
    msgtouser(unum,"Couldn't read database - try again later.");
    return;
  }
  if (ad.authlevel>=getauthlevel(unum)) {
    msgtouser(unum,"You cannot delete an user with a higher or equal authlevel");
    return;
  }
  mysql_escape_string(tmps2,ad.username,strlen(ad.username));
  sprintf(tmps4,"DELETE FROM users WHERE username='%s'",tmps2);
  res=mysql_query(&sqlcon,tmps4);
  if (res==0) {
    msgtouser(unum,"User deleted.");
  } else {
    msgtouser(unum,"Query to database failed.");
  }
}

void donoticeset(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; authdata ad; autheduser *au;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: noticeme on|off");
    return;
  }
  if (getauthlevel(unum)<1) {
    msgtouser(unum,"This command only makes sense when you are authed");
    return;
  }
  toLowerCase(tmps3);
  if ((strcmp(tmps3,"on")!=0) && (strcmp(tmps3,"off")!=0)) {
    msgtouser(unum,"Syntax: noticeme on|off");
    return;
  }
  getauthedas(tmps2,unum);
  getuserdata(&ad,tmps2);
  if (ad.authlevel<1) {
    msgtouser(unum,"That user does not exist in the database or database is down.");
    return;
  }
  if (strcmp(tmps3,"on")==0) { ad.wantsnotice=1; }
  if (strcmp(tmps3,"off")==0) { ad.wantsnotice=0; }
  updateuserinul(ad);
  au=getalentry(unum);
  au->wantsnotice=ad.wantsnotice;
  msgtouser(unum,"OK.");
}

void dobroadcast(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; userdata *a; int i;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %[^\n]",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: broadcast message");
    return;
  }
  for (i=0;i<SIZEOFUL;i++) {
    a=uls[i];
    while (a!=NULL) {
      longtotoken(a->numeric,tmps2,5);
      fprintf(sockout,"%sAAB O %s :(Broadcast) %s\r\n",servernumeric,tmps2,tmps3);
      a=(void *)a->next;
    }
  }
  fflush(sockout);
}

void dosbroadcast(long unum, char *tail) {
  int res, i; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  userdata *a; long targnum;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %[^\n]",tmps2,tmps4,tmps3);
  if (res!=3) {
    msgtouser(unum,"Syntax: sbroadcast targetserver message");
    return;
  }
  targnum=getsrvnum(tmps4);
  if (targnum<0) {
    msgtouser(unum,"That server is not on the network.");
    return;
  }
  res=0;
  for (i=0;i<SIZEOFUL;i++) {
    a=uls[i];
    while (a!=NULL) {
      if ((a->numeric>>SRVSHIFT)==targnum) {
        longtotoken(a->numeric,tmps2,5);
        fprintf(sockout,"%sAAB O %s :(SBroadcast) %s\r\n",servernumeric,tmps2,tmps3);
        res++;
      }
      a=(void *)a->next;
    }
  }
  fflush(sockout);
  sprintf(tmps2,"Message sent to %d Users.",res);
  msgtouser(unum,tmps2);
}

void dosyncglines(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE]; int i=0;
  MYSQL_RES *myres; MYSQL_ROW myrow;
  if (!checkauthlevel(unum,900)) { return; }
  purgeglines();
  res=mysql_query(&sqlcon,"SELECT * FROM glines");
  if (res!=0) {
    sprintf(logline,"!!! Failed to query gline-list from database !!!"); putlog();
    msgtouser(unum,logline);
    return;
  }
  myres=mysql_store_result(&sqlcon);
  if (mysql_num_fields(myres)!=4) {
    sprintf(logline,"!!! GList in database has invalid format !!!"); putlog();
    msgtouser(unum,logline);
    return;
  }
  while ((myrow=mysql_fetch_row(myres))) {
    fprintf(sockout,"%s GL * +%s %ld :%s\r\n",servernumeric,myrow[0],atol(myrow[2])-getnettime(),myrow[3]);
    i++;
  }
  fflush(sockout);
  sprintf(tmps2,"%d GLINEs resynched.",i);
  msgtouser(unum,tmps2);
  mysql_free_result(myres);
}

void dosubnetlist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; char *a;
  int x1,y1,z1; userdata *b; int c=0; int x2, y2, z2, w2; int i;
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: subnetlist x.y.z");
    return;
  }
  a=strtok(tmps3,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps2,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps4,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps5,a);
  x1=atoi(tmps2); y1=atoi(tmps4); z1=atoi(tmps5);
  sprintf(tmps2,"Clients on %d.%d.%d.*:",x1,y1,z1);
  msgtouser(unum,tmps2);
  for (i=0;i<SIZEOFUL;i++) {
    b=uls[i];
    while (b!=NULL) {
      splitip(b->realip,&x2,&y2,&z2,&w2);
      if ((x1==x2) && (y1==y2) && (z1==z2)) {
        c++;
        sprintf(tmps3,"%s!%s@%s",b->nick,b->ident,b->host);
        sprintf(tmps2,"%-50s %3d.%3d.%3d.%3d %s",tmps3,x2,y2,z2,w2,b->realname);
        msgtouser(unum,tmps2);
      }
      b=(void *)b->next;
    }
  }
  sprintf(tmps2,"--- End of list. %d matches. ---",c);
  msgtouser(unum,tmps2);
}

void dolistauthed(long unum, char *tail) {
  autheduser *a; char tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  a=als;
  sprintf(tmps4,"%-50s %-20s %-4s","Who","Authed as","Lev");
  msgtouser(unum,tmps4);
  while (a!=NULL) {
    numtohostmask(a->numeric,tmps3);
    sprintf(tmps4,"%-50s %-20s %4d",tmps3,a->authedas,a->authlevel);
    msgtouser(unum,tmps4);
    a=(void *)a->next;
  }
  msgtouser(unum,"End of authlist");
}

void dosave(long unum, char *tail) {
  if (!checkauthlevel(unum,900)) { return; }
  saveall();
  msgtouser(unum,"Done. Tried to save everything.");
}

void dosntrustadd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE], tmps7[TMPSSIZE]; char *a;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps6,tmps7);
  if (res!=5) {
    msgtouser(unum,"Syntax: sntrustadd X.Y.Z numusers duration reason");
    msgtouser(unum,"Example: sntrustadd 127.0.0 500 1d LAN-Party");
    return;
  }
  if (atoi(tmps4)==0) { msgtouser(unum,"not a valid number for numusers"); return; }
  if (durationtolong(tmps6)<=0) { msgtouser(unum,"not a valid duration"); return; }
  if (strlen(tmps7)>=200) { tmps7[199]='\0'; }
  res=atoi(tmps4);
  a=strtok(tmps3,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps2,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps4,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps5,a);
  snltrustadd(atoi(tmps2),atoi(tmps4),atoi(tmps5),res,getnettime()+durationtolong(tmps6),tmps7);
  sprintf(tmps3,"Subnet %d.%d.%d.* now trusted for %d users (reason: %s)",atoi(tmps2),atoi(tmps4),atoi(tmps5),res,tmps7);
  msgtouser(unum,tmps3);
}

void dosntrustdel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; char *a;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: sntrustdel X.Y.Z");
    msgtouser(unum,"Example: sntrustdel 127.0.0");
    return;
  }
  a=strtok(tmps3,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps2,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps4,a);
  a=strtok(NULL,".");
  if (a==NULL) { msgtouser(unum,"Illegal subnet"); return; }
  strcpy(tmps5,a);
  res=snltrustdel(atoi(tmps2),atoi(tmps4),atoi(tmps5));
  sprintf(tmps3,"%d.%d.%d.*",atoi(tmps2),atoi(tmps4),atoi(tmps5));
  if (res!=1) {
    sprintf(tmps2,"Subnet %s was not trusted.",tmps3);
  } else {
    sprintf(tmps2,"Deleted trustentry for %s.",tmps3);
  }
  msgtouser(unum,tmps2);
}

void dosntrustlist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE]; int i;
  trustedsubnet *a;
  if (!checkauthlevel(unum,900)) { return; }
  trustlistexpire();
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if ((res!=1) && (res!=2)) {
    msgtouser(unum,"Syntax: sntrustlist [pattern]");
    return;
  }
  if (res==1) { strcpy(tmps3,"*"); }
  res=0;
  sprintf(tmps2,"Trustlist-Entrys that match %s:",tmps3);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"%-13s %-3s %-20s %s","Net","n","expires in","Reason");
  msgtouser(unum,tmps2);
  a=tsn;
  for (i=0;i<tsncu;i++) {
    sprintf(tmps4,"%d.%d.%d",a->a,a->b,a->c);
    strcpy(tmps5,tmps3);
    if (match2strings(tmps5,tmps4)) {
      longtoduration(tmps6,a->expires-getnettime());
      sprintf(tmps2,"%-13s %3d %-20s %s",tmps4,a->n,tmps6,a->reason);
      msgtouser(unum,tmps2);
      res++;
    }
    a++;
  }
  sprintf(tmps2,"--- End of list. %d matches ---",res);
  msgtouser(unum,tmps2);
}

void dospew(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE]; int i; int searchtype; int j;
  userdata *a; channel **d; int chanmodes;
  if (!checkauthlevel(unum,600)) { return; }
  res=sscanf(tail,"%s %[^\n]",tmps2,tmps3);
  if (res<2) {
    msgtouser(unum,"Syntax: spew host|searchmask");
    msgtouser(unum,"anything that does not contain wildcards is treated as a host,");
    msgtouser(unum,"anything that does as a searchmask that is matched against 'realname nick!user@host'");
    return;
  }
  toLowerCase(tmps3);
  searchtype=0; // type: host
  for (i=0;i<strlen(tmps3);i++) {
    if ((tmps3[i]=='*') || (tmps3[i]=='?') || (tmps3[i]=='\\')) { searchtype=1; }
  }
  unescapestring(tmps3,tmps2);
  strcpy(tmps3,tmps2);
  if (searchtype==1) {
    sprintf(tmps4,"Users matching %s",tmps3);
  } else {
    sprintf(tmps4,"Users on %s",tmps3);
  }
  msgtouser(unum,tmps4);
  j=0;
  for (i=0;i<SIZEOFUL;i++) {
    a=uls[i];
    if (j>500) { break; }
    while (a!=NULL) {
      if (j>500) { break; }
      strcpy(tmps5,tmps3);
      if (searchtype==1) {
        sprintf(tmps2,"%s %s!%s@%s",a->realname,a->nick,a->ident,a->host);
      } else {
        strcpy(tmps2,a->host);
      }
      toLowerCase(tmps2);
      if (match2strings(tmps5,tmps2)) {
        int iamtired=0;
        j++;
        sprintf(tmps4,"%s!%s@%s (%s)",a->nick,a->ident,a->host,a->realname);
        msgtouser(unum,tmps4);
        if (a->chans.cursi>0) {
          if (a->chans.cursi==1) {
            strcpy(tmps4,"On channel:");
          } else {
            strcpy(tmps4,"On channels:");
          }
          for (iamtired=0;iamtired<(a->chans.cursi);iamtired++) {
            d=(channel **)(a->chans.content);
            if ((strlen(d[iamtired]->name)+strlen(tmps4))>400) {
              strcat(tmps4," [...]"); break;
            } else {
              chanmodes=getchanmode2(d[iamtired],a->numeric);
              strcat(tmps4," (");
              if (chanmodes<0) {
                strcat(tmps4,"??");
              } else {
                if (isflagset(chanmodes,um_o)) { strcat(tmps4,"@"); } else { strcat(tmps4,"_"); }
                if (isflagset(chanmodes,um_v)) { strcat(tmps4,"+"); } else { strcat(tmps4,"_"); }
              }
              strcat(tmps4,")");
              strcat(tmps4,d[iamtired]->name);
            }
          }
        } else {
          strcpy(tmps4,"Not on any (nonlocal-)channels");
        }
        msgtouser(unum,tmps4);
      }
      a=(void *)a->next;
    }
  }
  if (j>500) { msgtouser(unum,"-- More than 500 matches - list truncated"); }
  sprintf(tmps4,"-- End of list -- Found %d user%s",j,(j==1) ? "" : "s");
  msgtouser(unum,tmps4);
}

int rnlcomp(rnlentry *a, rnlentry *b) {
  if (a->n==b->n) { return 0; }
  if (a->n>b->n) { return -1; } else { return 1; }
}

void rnlsort(array *myrnl) {
  if (myrnl->cursi==0) { return; }
  qsort(myrnl->content,myrnl->cursi,sizeof(rnlentry),(void *)rnlcomp);
}

void dornc(long unum, char *tail) {
  char tmps2[TMPSSIZE]; int i=0; rnlentry *tmpp; int j, k;
  array myrnl; rnlentry *tmp2; long l;
  msgtouser(unum,"Top 20 Realnames on the net:");
  array_init(&myrnl,sizeof(rnlentry));
  for (j=0;j<256;j++) {
    tmp2=(rnlentry *)(mainrnl[j].content);
    for (k=0;k<mainrnl[j].cursi;k++) {
      l=array_getfreeslot(&myrnl);
      tmpp=(rnlentry *)(myrnl.content);
      tmpp[l]=tmp2[k];
    }
  }
  rnlsort(&myrnl);
  tmpp=(rnlentry *)(myrnl.content);
  while ((i<20) && (i<myrnl.cursi)) {
    sprintf(tmps2,"(%2d) %3d %-50s",i+1,tmpp[i].n,tmpp[i].r);
    msgtouser(unum,tmps2);
    i++;
  }
  array_free(&myrnl);
}

void dointtobase64(long unum, char *tail) {
  unsigned long a; int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  res=sscanf(tail,"%s %lu",tmps2,&a);
  if (res!=2) {
    msgtouser(unum,"Syntax: inttobase64 integer");
    return;
  }
  longtotoken(a,tmps2,10);
  sprintf(tmps3,"%lu (10) == %s (BASE64)",a,tmps2);
  msgtouser(unum,tmps3);
}

void dobase64toint(long unum, char *tail) {
  unsigned long a; int res; static char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res!=2) {
    msgtouser(unum,"Syntax: base64toint base64encodednumber");
    return;
  }
  a=tokentolong(tmps3);
  sprintf(tmps2,"%s (BASE64) == %lu (10)",tmps3,a);
  msgtouser(unum,tmps2);
}

void dodie(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  if (!checkauthlevel(unum,998)) { return; }
  res=sscanf(tail,"%s %[^\n]",tmps2,tmps3);
  if (res<2) { numtonick(unum,tmps2); sprintf(tmps3,"Terminated by %s",tmps2); }
  msgtouser(unum,"Terminating...");
  fprintf(sockout,"%s SQ %s :%s\r\n",servernumeric,myname,tmps3);
  fflush(sockout);
}

void donoticemask(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; authdata ad; unsigned long ul;
  if (getauthlevel(unum)<1) {
    msgtouser(unum,"This command only makes sense when you are authed.");
    return;
  }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res<2) { /* Show noticemask */
    if (!isircop(unum)) {
      msgtouser(unum,"Warning: You are not opered, so whatever your noticemask is set to, it will be 0 as long as you're not opered.");
    }
    ul=getnoticemask(unum);
    sprintf(tmps2,"Your current noticemask is %lu",ul);
    msgtouser(unum,tmps2);
  } else { /* Set new mask */
    getauthedas(tmps2,unum);
    getuserdata(&ad,tmps2);
    if (ad.authlevel<1) {
      msgtouser(unum,"That user does not exist in the database or database is down.");
      return;
    }
    ul=strtoul(tmps3,NULL,10);
    ad.noticemask=ul;
    updateuserinul(ad);
    setnoticemask(unum,ul);
    sprintf(tmps2,"Noticemask set to %lu",ul);
    msgtouser(unum,tmps2);
  }
}
